"""
Topology Extractor
==================

#. :class:`.TopologyExtractor`

Class for defining a topology from a molecule and disconnections.

"""

import logging

from ..networkx import Network
from .topology_info import TopologyInfo

logger = logging.getLogger(__name__)


class TopologyExtractor:
    """
    Extractor of topology definitions from a molecule.

    Examples
    --------

    Using a SMARTS string and the
    :class:`stk.SmartsFunctionalGroupFactory`, you can split a molecule
    at any point to define a topology.

    .. code-block:: python

        import stk
        import stko

        smarts = '[#6]~[#7]'

        mol = stk.BuildingBlock(
            smiles='C1=CC=C(C=C1)CN',
            functional_groups=(stk.SmartsFunctionalGroupFactory(
                smarts=smarts,
                bonders=(),
                deleters=(),
            ), )
        )

        broken_bonds_by_id = []
        disconnectors = []
        for fg in mol.get_functional_groups():
            print(fg)
            atom_ids = list(fg.get_atom_ids())
            bond_c = atom_ids[0]
            bond_n = atom_ids[1]
            broken_bonds_by_id.append(sorted((bond_c, bond_n)))
            disconnectors.extend((bond_c, bond_n))

        print(broken_bonds_by_id)
        print(disconnectors)
        print('--')

        new_topology_graph = stko.TopologyExtractor()
        tg_info = new_topology_graph.extract_topology(
            molecule=mol,
            broken_bonds_by_id=broken_bonds_by_id,
            disconnectors=set(disconnectors),
        )
        print(tg_info.get_vertex_positions())
        print(tg_info.get_connectivities())
        print(tg_info.get_edge_pairs())

    """

    def extract_topology(
        self,
        molecule,
        broken_bonds_by_id,
        disconnectors,
    ):
        """
        Extract a toplogy defining a molecule with disconnections.

        Parameters
        ----------
        molecule : :class:`stk.Molecule`
            Molecule to get underlying topology of.

        broken_bonds_by_id : :class:`iterable` of :class:`tuple`
            Tuples of bonds to break by atom id.

        disconnectors : :class:`set`
            Atom ids of disconnection points.

        Returns
        -------
        :class:`.TopologyInfo`
            Information of the underlying topology.

        """

        connected_graphs = self.get_connected_graphs(
            molecule=molecule,
            atom_ids_to_disconnect=broken_bonds_by_id,
        )

        centroids = {}
        connectivities = {}
        edge_pairs = []
        for i, cg in enumerate(connected_graphs):
            centroids[i] = molecule.get_centroid(
                atom_ids=[i.get_id() for i in cg]
            )
            disconnections = 0
            for atom in cg:
                if atom.get_id() in disconnectors:
                    disconnections += 1
            connectivities[i] = disconnections

        for pair in broken_bonds_by_id:
            a1_g, a2_g = (
                i for i, cg in enumerate(connected_graphs)
                if pair[0] in set(atom.get_id() for atom in cg)
                or pair[1] in set(atom.get_id() for atom in cg)
            )
            edge_pairs.append((a1_g, a2_g))

        return TopologyInfo(centroids, connectivities, edge_pairs)

    def get_connected_graphs(
        self,
        molecule,
        atom_ids_to_disconnect
    ):
        graph = Network.init_from_molecule(molecule)
        graph = graph.with_deleted_bonds(atom_ids_to_disconnect)
        connected_graphs = graph.get_connected_components()
        return connected_graphs

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return f'<{self.__class__.__name__} at {id(self)}>'



class CaseData:
    """
    An :class:`.Du` test case.

    Attributes
    ----------
    atom : :class:`.Du`
        The atom being tested.

    id : :class:`int`
        The correct id.

    """

    def __init__(self, atom, id):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        atom : :class:`.Du`
            The atom being tested.

        id : :class:`int`
            The correct id.

        """

        self.atom = atom
        self.id = id



import pytest
import numpy as np
import stko
import stk


def a_molecule():
    return stk.BuildingBlock(smiles='CC')


@pytest.fixture
def unoptimized_mol():
    return a_molecule()


class PassingOptimizer(stko.Optimizer):

    def optimize(self, mol):
        return a_molecule().with_centroid(np.array(([1, 3, 3])))


class FailingOptimizer(stko.Optimizer):

    def optimize(self, mol):
        raise Exception()


@pytest.fixture
def passing_optimizer():
    return PassingOptimizer()


@pytest.fixture
def failing_optimizer():
    return FailingOptimizer()



import pytest
import numpy as np
import os
from stko import GulpUFFOptimizer, GulpUFFMDOptimizer, get_metal_atoms
from .conftest import a_molecule


class FakeGulpUFFOptimizer(GulpUFFOptimizer):

    def optimize(self, mol):
        return a_molecule().with_centroid(np.array(([1, 3, 3])))


class FakeGulpUFFMDOptimizer(GulpUFFMDOptimizer):

    def optimize(self, mol):
        return a_molecule().with_centroid(np.array(([1, 3, 3])))


@pytest.fixture
def position_section():
    return (
        '\ncartesian\n'
        'C1 core -0.74031 0.03171 0.10194\n'
        'C1 core 0.75974 -0.01804 -0.03434\n'
        'H1 core -1.14779 -0.63142 -0.70939\n'
        'H1 core -1.11274 -0.39237 1.04655\n'
        'H1 core -1.12625 1.0535 -0.10345\n'
        'H1 core 0.97939 0.03507 -1.13219\n'
        'H1 core 1.25573 0.87341 0.40438\n'
        'H1 core 1.13222 -0.95187 0.4265\n'
    )


@pytest.fixture
def bond_section():
    return (
        '\nconnect 1 2 \n'
        'connect 1 3 \n'
        'connect 1 4 \n'
        'connect 1 5 \n'
        'connect 2 6 \n'
        'connect 2 7 \n'
        'connect 2 8 \n'
    )


@pytest.fixture
def species_section():
    return '\nspecies\nC1 C_3\nH1 H_\n'


def test_gulp_position_section(unoptimized_mol, position_section):
    opt = FakeGulpUFFOptimizer(
        gulp_path='',
        maxcyc=1000,
        metal_FF=None,
        metal_ligand_bond_order=None,
        conjugate_gradient=False,
        output_dir='',
    )
    opt.assign_FF(unoptimized_mol)
    type_translator = opt._type_translator()
    test = opt._position_section(unoptimized_mol, type_translator)
    assert position_section == test


def test_gulp_bond_section(unoptimized_mol, bond_section):
    opt = FakeGulpUFFOptimizer(
        gulp_path='',
    )
    opt.assign_FF(unoptimized_mol)
    metal_atoms = get_metal_atoms(unoptimized_mol)
    test = opt._bond_section(unoptimized_mol, metal_atoms)
    assert bond_section == test


def test_gulp_species_section(unoptimized_mol, species_section):
    opt = FakeGulpUFFOptimizer(
        gulp_path='',
    )
    opt.assign_FF(unoptimized_mol)
    type_translator = opt._type_translator()
    test = opt._species_section(type_translator)
    assert species_section == test


@pytest.fixture
def atom_types():
    return ['C', 'C', 'H', 'H', 'H', 'H', 'H', 'H']


@pytest.fixture
def trajectory():
    """
    Defines output of the trajectory properties ignoring coords.

    """

    return {
        0: {
            'time': 2.99999999999978,
            'KE': 0.148961461675797,
            'E': 0.322914341829668,
            'T': 192.069356724061,
        },
        1: {
            'time': 5.0,
            'KE': 0.207467138253947,
            'E': 0.245970616427967,
            'T': 267.505966560289,
        },
        2: {
            'time': 7.00000000000067,
            'KE': 0.267980085285543,
            'E': 0.207888721719671,
            'T': 345.530730006366,
        },
        3: {
            'time': 9.00000000000045,
            'KE': 0.203650341841703,
            'E': 0.203495426725446,
            'T': 262.584629031782,
        },
        4: {
            'time': 10.9999999999993,
            'KE': 0.209627546107803,
            'E': 0.246948965763409,
            'T': 270.291573938759,
        },
    }


@pytest.fixture
def min_energy_time_step():
    return 3


@pytest.fixture
def xyz_string():
    return (
        '8\n'
        '1,5.0,0.207467138253947,0.245970616427967,267.505966560289\n'
        'C -0.74873 0.01254 0.08187\n'
        'C 0.75325 -0.00453 -0.03139\n'
        'H -1.21565 -0.9294 -0.08591\n'
        'H -1.00574 0.41323 1.09345\n'
        'H -1.09608 0.79503 -0.60147\n'
        'H 1.02245 -0.86479 -0.73236\n'
        'H 1.19827 0.93931 -0.43282\n'
        'H 1.25464 -0.29976 0.89505\n'
    )


def test_gulp_convert_traj_to_xyz(atom_types, trajectory):
    test_dir = os.path.dirname(os.path.abspath(__file__))
    test_xyz = f'{test_dir}/fixtures/gulp_MD_template.xyz'
    test_traj = f'{test_dir}/fixtures/gulp_MD.trg'
    opt = FakeGulpUFFMDOptimizer(
        gulp_path='',
    )
    test_atom_types, test_trajectory_data, _ = (
        opt._convert_traj_to_xyz(
            output_xyz=test_xyz,
            output_traj=test_traj,
        )
    )

    for i, t in zip(atom_types, test_atom_types):
        assert i == t

    for ts in trajectory:
        test_ts_dict = test_trajectory_data[ts]
        ts_dict = trajectory[ts]
        assert test_ts_dict['time'] == ts_dict['time']
        assert test_ts_dict['KE'] == ts_dict['KE']
        assert test_ts_dict['E'] == ts_dict['E']
        assert test_ts_dict['T'] == ts_dict['T']


def test_gulp_calculate_lowest_energy_conformer(min_energy_time_step):
    test_dir = os.path.dirname(os.path.abspath(__file__))
    test_xyz = f'{test_dir}/fixtures/gulp_MD_template.xyz'
    test_traj = f'{test_dir}/fixtures/gulp_MD.trg'
    opt = FakeGulpUFFMDOptimizer(
        gulp_path='',
    )
    atom_types, trajectory_data, xyz_traj_lines = (
        opt._convert_traj_to_xyz(
            output_xyz=test_xyz,
            output_traj=test_traj,
        )
    )
    min_ts = opt._calculate_lowest_energy_conformer(trajectory_data)

    assert min_ts == min_energy_time_step


def test_gulp_write_conformer_xyz_file(xyz_string):
    test_dir = os.path.dirname(os.path.abspath(__file__))
    test_xyz = f'{test_dir}/fixtures/gulp_MD_template.xyz'
    test_traj = f'{test_dir}/fixtures/gulp_MD.trg'
    opt = FakeGulpUFFMDOptimizer(
        gulp_path='',
    )
    atom_types, trajectory_data, xyz_traj_lines = (
        opt._convert_traj_to_xyz(
            output_xyz=test_xyz,
            output_traj=test_traj,
        )
    )

    test_output = f'{test_dir}/fixtures/conformer.xyz'
    opt._write_conformer_xyz_file(
        ts=1,
        ts_data=trajectory_data[1],
        filename=test_output,
        atom_types=atom_types,
    )

    test_string = ''
    with open(test_output, 'r') as f:
        for line in f.readlines():
            print(line)
            test_string += line

    assert test_string == xyz_string




class CaseData:
    """
    A test case.

    Attributes
    ----------
    molecule : :class:`stk.Molecule`
        The molecule to test.

    zmatrix : :class:`str`
        The correct Z-matrix string.

    """

    def __init__(self, molecule, zmatrix):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        molecule : :class:`stk.Molecule`
            The molecule to test.

        zmatrix : :class:`str`
            The correct Z-matrix string.

        """

        self.molecule = molecule
        self.zmatrix = zmatrix



def test_get_atomic_number(case_data):
    """
    Test :meth:`.Du.get_atomic_number`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case, containing the atom to test and its correct atomic
        number.

    Returns
    -------
    None : :class:`NoneType`

    """

    assert case_data.atom.get_atomic_number() == 0



import pytest
import stk


_pd = stk.BuildingBlock(
    smiles='[Pd+2]',
    functional_groups=(
        stk.SingleAtom(stk.Pd(0, charge=2))
        for i in range(4)
    ),
    position_matrix=[[0, 0, 0]],
)

# Define a bidentate ligand with two functional groups.
_bidentate_ligand = stk.BuildingBlock(
    smiles='NCCN',
    functional_groups=[
        stk.SmartsFunctionalGroupFactory(
            smarts='[#7]~[#6]',
            bonders=(0, ),
            deleters=(),
        ),
    ],
)

# Construct a cis-protected square planar metal complex.
_complex = stk.ConstructedMolecule(
    topology_graph=stk.metal_complex.CisProtectedSquarePlanar(
        metals=_pd,
        ligands=_bidentate_ligand,
    ),
)


@pytest.fixture(
    params=(
        stk.BuildingBlock('NCCN'),
        stk.BuildingBlock('c1ccccc1'),
        _complex,
    ),
)
def molecule(request):
    """
    A :class:`.CaseData` instance.
    """

    return request.param



"""
Unit Cell
=========

Class holding periodic cell information.

"""

import numpy as np
import logging
from stk import PeriodicInfo

from .utilities import get_from_parameters

logger = logging.getLogger(__name__)


class UnitCell(PeriodicInfo):
    """
    Unit cell information for periodic systems.

    """

    @classmethod
    def _update_periodic_info(cls, vector_1, vector_2, vector_3):
        """
        Return clone of :class:`.UnitCell` with new parameters.

        """

        clone = cls.__new__(cls)
        UnitCell.__init__(
            self=clone,
            vector_1=vector_1,
            vector_2=vector_2,
            vector_3=vector_3,
        )

        return clone

    def with_cell_from_vectors(self, vector_1, vector_2, vector_3):
        """
        Update cell.

        Parameters
        ----------
        vector_1 : :class:`numpy.ndarray`
            First cell lattice vector of shape (3, ) in
            Angstrom.

        vector_2 : :class:`numpy.ndarray`
            Second cell lattice vector of shape (3, ) in
            Angstrom.

        vector_3 : :class:`numpy.ndarray`
            Third cell lattice vector of shape (3, ) in
            Angstrom.

        Returns
        -------
        :class:`.UnitCell`
            Clone with updated cell parameters.

        """

        return self.__class__._update_periodic_info(
            vector_1=vector_1,
            vector_2=vector_2,
            vector_3=vector_3,
        )

    def with_cell_from_turbomole(self, filename):
        """
        Update cell from structure in Turbomole coord file.

        Returns
        -------
        :class:`.UnitCell`
            Clone with updated cell parameters.

        """

        bohr_to_ang = 0.5291772105638411

        with open(filename, 'r') as f:
            content = f.readlines()

        periodicity = False
        lattice_vectors = None
        lattice_units = None
        cell_parameters = None
        cell_units = None
        for line_number, line in enumerate(content):
            if '$periodic' in line:
                periodicity = int(line.rstrip().split()[1])
            if '$cell' in line:
                if 'angs' in line:
                    cell_units = 'angstrom'
                elif 'bohr' in line:
                    cell_units = 'bohr'
                else:
                    raise ValueError('cell not in Angstroms.')
                cell_parameters = [
                    float(j)
                    for j in content[line_number+1].rstrip().split()
                ]
            if '$lattice' in line:
                if 'angs' in line:
                    lattice_units = 'angstrom'
                elif 'bohr' in line:
                    lattice_units = 'bohr'
                else:
                    raise ValueError('lattice not in Angstroms.')
                lattice_vectors = (
                    np.array([
                        float(j) for j
                        in content[line_number+1].rstrip().split()
                    ]),
                    np.array([
                        float(j) for j
                        in content[line_number+2].rstrip().split()
                    ]),
                    np.array([
                        float(j) for j
                        in content[line_number+3].rstrip().split()
                    ]),
                )

        # Check that cell is only defined once.
        chk2 = (
            lattice_vectors is not None and cell_parameters is not None
        )
        if periodicity and chk2:
            raise RuntimeError(
                'The cell is defined twice in the file.'
            )

        if lattice_vectors is not None:
            vector_1 = (
                lattice_vectors[0]*bohr_to_ang
                if lattice_units == 'bohr' else lattice_vectors[0]
            )
            vector_2 = (
                lattice_vectors[1]*bohr_to_ang
                if lattice_units == 'bohr' else lattice_vectors[0]
            )
            vector_3 = (
                lattice_vectors[2]*bohr_to_ang
                if lattice_units == 'bohr' else lattice_vectors[0]
            )
        elif cell_parameters is not None:
            vector_1, vector_2, vector_3 = get_from_parameters(
                a=(
                    cell_parameters[0] * bohr_to_ang
                    if cell_units == 'bohr'
                    else cell_parameters[0]
                ),
                b=(
                    cell_parameters[1] * bohr_to_ang
                    if cell_units == 'bohr'
                    else cell_parameters[1]
                ),
                c=(
                    cell_parameters[2] * bohr_to_ang
                    if cell_units == 'bohr'
                    else cell_parameters[2]
                ),
                alpha=cell_parameters[3],
                beta=cell_parameters[4],
                gamma=cell_parameters[5],
            )
        else:
            raise RuntimeError(
                'The cell is not defined in the file.'
            )
        # Update the cell.
        return self._update_periodic_info(vector_1, vector_2, vector_3)

    def with_cell_from_cif(self, filename):
        """
        Update cell from structure in CIF.

        Returns
        -------
        :class:`NoneType`
            Clone with updated cell parameters.

        """

        cell_info = {}

        targets = {
            '_cell_length_a': 'a',
            '_cell_length_b': 'b',
            '_cell_length_c': 'c',
            '_cell_angle_alpha': 'alpha',
            '_cell_angle_beta': 'beta',
            '_cell_angle_gamma': 'gamma',
        }

        with open(filename, 'r') as f:
            lines = f.readlines()

        for targ in targets:
            for line in lines:
                # Avoid running through the rest.
                if targets[targ] in cell_info.keys():
                    break
                splits = line.rstrip().split(' ')
                if splits[0] == targ:
                    cell_info[targets[targ]] = float(splits[-1])

        vector_1, vector_2, vector_3 = get_from_parameters(
            a=cell_info['a'],
            b=cell_info['b'],
            c=cell_info['c'],
            alpha=cell_info['alpha'],
            beta=cell_info['beta'],
            gamma=cell_info['gamma'],
        )
        # Update the cell.
        return self._update_periodic_info(vector_1, vector_2, vector_3)



"""
Molecule Transformer
====================

#. :class:`.MoleculeTransformer`

Class for splitting a molecule into many with new connectors.

"""

import logging
from rdkit.Chem import AllChem as rdkit

import stk

from ..atoms import Du

logger = logging.getLogger(__name__)


class MoleculeTransformer:
    """
    Split an stk.molecule into many with new functional groups.

    Examples
    --------

    Given a molecule, this class allows you to cap a split molecule
    (see :class:`MoleculeSplitter`) at the broken bond with an atom
    defined in `replacer_smarts`.

    .. code-block:: python

        import stk
        import stko

        full_mol = stk.BuildingBlock('C1=CC=NC(=C1)C=NC2=CC=C(C=C2)Br')

        splitter = stko.MoleculeSplitter(
            breaker_smarts='[#6X3]~[#7X2]~[#6X3H1]~[#6X3!H1]',
            bond_deleter_ids=(0, 1),
        )
        split_mols = splitter.split(full_mol)

        transformer = stko.MoleculeTransformer(
            replacer_smarts='[Br]',
            functional_groups=(stk.BromoFactory(), ),
        )
        for split in split_mols:
            transformed_mol = transformer.transform(split)

    """

    def __init__(
        self,
        replacer_smarts,
        functional_groups,
    ):
        """
        Initialize a :class:`.MoleculeTransformer`.

        Parameters
        ----------
        replacer_smarts : :class:`str`
            SMARTS string of atom to replace dummy atoms with. This
            must be a single atom.

        functional_groups : :class:`iterable`
        of :class:`stk.FunctionalGroupFactory`
            Functional group factories to use to define new building
            block.

        Raises
        ------
        :class:`ValueError`
            If `replacer_smarts` does not correspond to a single atom.

        """

        # Add replacers bonded to * atoms.
        _atom_list = list(
            rdkit.MolFromSmarts(replacer_smarts).GetAtoms()
        )
        if len(_atom_list) != 1:
            raise ValueError(
                f'{replacer_smarts} corresponds {len(_atom_list)} '
                'atoms. Should be 1.'
            )

        for i in _atom_list:
            self._replacer = stk.Atom(
                id=0,
                atomic_number=i.GetAtomicNum(),
                charge=0,
            )

        self._functional_groups = functional_groups

    def transform(self, molecule):
        """
        Transform a molecule.

        Parameters
        ----------
        molecule : :class:`stk.Molecule`
            Molecule to modify.

        Returns
        -------
        molecule : :class:`stk.BuildingBlock`
            The resulting molecule.

        """

        rdkit_mol = molecule.to_rdkit_mol()
        rdkit.SanitizeMol(rdkit_mol)

        atoms = []
        for a in molecule.get_atoms():
            if isinstance(a, Du):
                atoms.append(stk.Atom(
                    id=a.get_id(),
                    atomic_number=self._replacer.get_atomic_number(),
                    charge=self._replacer.get_charge(),
                ))
            else:
                atoms.append(a)
        atoms = tuple(atoms)

        bonds = tuple(
            stk.Bond(
                atom1=atoms[b.get_atom1().get_id()],
                atom2=atoms[b.get_atom2().get_id()],
                order=b.get_order(),
            )
            for b in molecule.get_bonds()
        )
        position_matrix = molecule.get_position_matrix()
        building_block = stk.BuildingBlock.init(
            atoms=atoms,
            bonds=bonds,
            position_matrix=position_matrix,
        )
        building_block = stk.BuildingBlock.init_from_molecule(
            molecule=building_block,
            functional_groups=self._functional_groups,
        )

        return building_block



from ..utilities import is_equivalent_atom


def test_clone(positioned_atom):
    """
    Test :meth:`.PositionedAtom.clone`.

    Parameters
    ----------
    atom : :class:`.PositionedAtom`
        The atom to be cloned.

    Returns
    -------
    None : :class:`NoneType`

    """

    clone = positioned_atom.clone()
    is_equivalent_atom(positioned_atom, clone)



"""
OpenBabel Optimizers
================

#. :class:`.OpenBabel`

Wrappers for optimizers within the `openbabel` code.

"""

import logging
import os
import numpy as np
from openbabel import openbabel

from .optimizers import Optimizer

logger = logging.getLogger(__name__)


class OpenBabelError(Exception):
    ...


class ForceFieldSetupError(OpenBabelError):
    ...


class OpenBabel(Optimizer):
    """
    Use OpenBabel to optimize molecules with forcefields.[1]_

    Warning: this optimizer seems to be machine dependant, producing
    different energies after optimisation on Ubunut 18 vs. Ubuntu 20.

    Examples
    --------
    .. code-block:: python

        import stk
        import stko

        mol = stk.BuildingBlock('NCCNCCN')
        openbabel = stko.OpenBabel('uff')
        mol = openbabel.optimize(mol)

    References
    ----------
    .. [1] http://openbabel.org/dev-api/classOpenBabel_1_
    1OBForceField.shtml#a2f2732698efde5c2f155bfac08fd9ded

    """

    def __init__(
        self,
        forcefield,
        repeat_steps=10,
        sd_steps=50,
        cg_steps=50,
    ):
        """
        Initialize `openbabel` forcefield energy calculation.

        Parameters
        ----------
        forcefield : :class:`str`
            Forcefield to use. Options include `uff`, `gaff`,
            `ghemical`, `mmff94`.

        repeat_steps : :class:`int`
            Number of optimisation steps. Each optimisation step
            contains `sd_steps` steepest descent and then `cg_steps`
            conjugate gradient runs.

        sd_steps : :class:`int`
            Number of steepest descent steps per optimisations.

        cg_steps : :class:`int`
            Number of conjugate gradient steps per optimisations.

        """

        self._forcefield = forcefield
        self._repeat_steps = repeat_steps
        self._sd_steps = sd_steps
        self._cg_steps = cg_steps

    def optimize(self, mol):
        """
        Optimize `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        Returns
        -------
        mol : :class:`.Molecule`
            The optimized molecule.

        """

        temp_file = 'temp.mol'
        mol.write(temp_file)
        obConversion = openbabel.OBConversion()
        obConversion.SetInFormat("mol")
        OBMol = openbabel.OBMol()
        obConversion.ReadFile(OBMol, temp_file)
        OBMol.PerceiveBondOrders()
        os.system('rm temp.mol')

        forcefield = openbabel.OBForceField.FindForceField(
            self._forcefield
        )
        outcome = forcefield.Setup(OBMol)
        if not outcome:
            raise ForceFieldSetupError(
                f"{self._forcefield} could not be setup for {mol}"
            )
        for step in range(self._repeat_steps):
            forcefield.SteepestDescent(self._sd_steps)
            forcefield.GetCoordinates(OBMol)
            forcefield.ConjugateGradients(self._cg_steps)
            forcefield.GetCoordinates(OBMol)

        position_matrix = []
        for atom in openbabel.OBMolAtomIter(OBMol):
            # get coordinates
            position_matrix.append(
                np.array([atom.GetX(), atom.GetY(), atom.GetZ()])
            )

        mol = mol.with_position_matrix(np.asarray(position_matrix))
        return mol



import pytest
from pytest_lazyfixture import lazy_fixture
import stko

from .case_data import CaseData


@pytest.fixture
def case_data_1(id):
    """
    A :class:`.CaseData` instance.

    """

    return CaseData(
        atom=stko.Du(id),
        id=id,
    )


@pytest.fixture(
    params=(
        lazy_fixture('case_data_1'),
    )
)
def case_data(request):
    """
    A :class:`.CaseData` instance.

    """

    return request.param


@pytest.fixture(
    params=[0, 3],
)
def id(request):
    """
    An atom id.

    """

    return request.param


@pytest.fixture
def dummy_atom(id):
    """
    An :class:`.Du` instance.

    """

    return stko.Du(id).clone()



import stko


def test_zmatrix(case_data):
    """
    Test :meth:`.Converter.get_zmatrix`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        The test case.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_zmatrix(case_data.molecule, case_data.zmatrix)


def _test_zmatrix(molecule, zmatrix):
    """
    Test :meth:`.ZMatrix.get_zmatrix`.

    Parameters
    ----------
    molecule : :class:`stk.Molecule`
        The molecule to test.

    zmatrix : :class:`str`
        The correct Z-matrix string.

    Returns
    -------
    None : :class:`NoneType`

    """

    result = stko.ZMatrix().get_zmatrix(molecule)
    # Printing is useful for debugging.
    print(result)
    assert result == zmatrix



import numpy as np
import itertools as it


def is_equivalent_atom(atom1, atom2):
    assert atom1.get_id() == atom2.get_id()
    assert atom1.get_charge() == atom2.get_charge()
    assert atom1.get_atomic_number() == atom2.get_atomic_number()
    assert atom1.__class__ is atom2.__class__


def is_equivalent_bond(bond1, bond2):
    assert bond1.__class__ is bond2.__class__
    assert bond1.get_order() == bond2.get_order()
    assert bond1.get_periodicity() == bond2.get_periodicity()
    is_equivalent_atom(bond1.get_atom1(), bond2.get_atom1())
    is_equivalent_atom(bond1.get_atom2(), bond2.get_atom2())


def is_equivalent_molecule(molecule1, molecule2):
    atoms = it.zip_longest(
        molecule1.get_atoms(),
        molecule2.get_atoms(),
    )
    for atom1, atom2 in atoms:
        is_equivalent_atom(atom1, atom2)

    bonds = it.zip_longest(
        molecule1.get_bonds(),
        molecule2.get_bonds(),
    )
    for bond1, bond2 in bonds:
        is_equivalent_bond(bond1, bond2)


def inequivalent_position_matrices(molecule1, molecule2):
    pos1 = molecule1.get_position_matrix()
    pos2 = molecule2.get_position_matrix()

    assert not np.allclose(pos1, pos2)



from .periodic import *  # noqa
from .topology_extractor import *  # noqa
from .networkx import *  # noqa
from .molecule_modifiers import *  # noqa
from .atoms import *  # noqa
from .conversion import *  # noqa



"""
Calculator
==========

"""

from abc import ABC


class Calculator(ABC):
    """
    An abstract base class for calculators.

    """

    def calculate(self, mol):
        """
        Perform calculation on `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The :class:`.Molecule` to calculate properties of.

        Yields
        ------
        :class:`function`
            The function to perform the calculation.

        """

        raise NotImplementedError()

    def get_results(self, mol):
        """
        Get results of calculation on `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The :class:`.Molecule` to calculate properties of.

        Returns
        -------
        :class:`.Results`
            The result of the calculation.

        """

        raise NotImplementedError()



from .calculators import *  # noqa
from .xtb_calculators import *  # noqa
from .rdkit_calculators import *  # noqa
from .torsion_calculators import *  # noqa
from .shape_calculators import *  # noqa
from .orca_calculators import *  # noqa
from .open_babel_calculators import *  # noqa



"""
Torsion Calculators
===================

#. :class:`.TorsionCalculator`
#. :class:`.ConstructedMoleculeTorsionCalculator`

Methods to extract torsions from a molecule or constructed molecule.

"""

import logging

from .calculators import Calculator
from .results import TorsionResults, ConstructedMoleculeTorsionResults
from rdkit.Chem import TorsionFingerprints
from ..molecular.torsion import Torsion
from ..utilities.utilities import get_atom_maps

logger = logging.getLogger(__name__)


class TorsionCalculator(Calculator):
    """
    Uses rdkit to extract all torsions in a molecule.

    Note that the rdkit [1]_ function we use only outputs
    one torsion for each rotatable bond. We use the
    `TorsionFingerprints.CalculateTorsionLists` method.

    Examples
    --------

    .. code-block:: python

        import stk
        import stko

        # Create a molecule whose torsions we want to know.
        mol1 = stk.BuildingBlock('CCCNCCCN')

        # Create the calculator.
        tc = stko.TorsionCalculator()

        # Extract the torsions.
        tc_results = tc.get_results(mol1)
        for t, ang in tc_results.get_torsion_angles():
            print(t, ang, t.get_atom_ids())

    References
    ----------
    .. [1] http://rdkit.org/docs/source/
    rdkit.Chem.TorsionFingerprints.html

    """

    def calculate(self, mol):
        yield tuple(
            Torsion(*mol.get_atoms(atoms[0]))
            for atoms, _ in (
                TorsionFingerprints.CalculateTorsionLists(
                    mol.to_rdkit_mol()
                )[0]
            )
        )

    def get_results(self, mol):
        """
        Calculate the torsions of `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The :class:`.Molecule` whose torsions are to be calculated.

        Returns
        -------
        :class:`.TorsionResults`
            The torsions of the molecule.

        """

        return TorsionResults(self.calculate(mol), mol)


class ConstructedMoleculeTorsionCalculator(TorsionCalculator):
    """
    Uses rdkit to extract all torsions in a molecule.

    Note that the rdkit [1]_ function we use only outputs
    one torsion for each rotatable bond. We use the
    `TorsionFingerprints.CalculateTorsionLists` method.

    Examples
    --------

    .. code-block:: python

        import stk
        import stko

        # Create a molecule whose energy we want to know.
        bb1 = stk.BuildingBlock('NCCNCCN', [stk.PrimaryAminoFactory()])
        bb2 = stk.BuildingBlock('O=CCCC=O', [stk.AldehydeFactory()])
        polymer = stk.ConstructedMolecule(
            stk.polymer.Linear(
                building_blocks=(bb1, bb2),
                repeating_unit="AB",
                orientations=[0, 0],
                num_repeating_units=1
            )
        )

        # Create the calculator.
        tc = stko.ConstructedMoleculeTorsionCalculator()

        # Extract the torsions.
        tc_results = tc.get_results(polymer)

        # Get information about torsions in building blocks and in the
        # ConstructedMolecule.
        for t in tc_results.get_torsion_infos():
            print(
                'c', t.get_torsion(),
                t.get_building_block(),
                t.get_building_block_id(),
                t.get_building_block_torsion(),
            )

    References
    ----------
    .. [1] http://rdkit.org/docs/source/
    rdkit.Chem.TorsionFingerprints.html

    """

    def get_results(self, mol):
        """
        Calculate the torsions of `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The :class:`.Molecule` whose torsions are to be calculated.

        Returns
        -------
        :class:`.TorsionResults`
            The torsions of the molecule.

        """

        return ConstructedMoleculeTorsionResults(
            generator=self.calculate(mol),
            mol=mol,
        )


class MatchedTorsionCalculator(ConstructedMoleculeTorsionCalculator):
    """
    Matches rdkit generated torsions with building block torsions.

    """

    def calculate(self, mol):
        """
        Extract torsions with rdkit, then match to building blocks.

        This method loops through each rdkit generated torsion. For
        each torsion, it checks if the two interior atoms of the
        torsion come from the two interior (central) atoms of some
        torsion in a building block. If so, it replaces the end atoms
        of the torsion with the atoms corresponding to the end atoms
        of the underlying building block torsion.

        Parameters
        ----------
        mol : :class:`.ConstructedMolecule`
            The :class:`.ConstructedMolecule` whose torsions are to be
            calculated.

        """

        torsions = list(next(super().calculate(mol)))
        atom_maps = get_atom_maps(mol)

        # Loop over torsions, updating each to match a building block
        # torsion if possible.
        for i, torsion in enumerate(torsions):
            # Atom ids, atom infos, and building block ids of current
            # torsion in constructed molecule.
            atom_ids = list(torsion.get_atom_ids())
            atom_infos = list(mol.get_atom_infos(atom_ids))
            build_block_ids = [
                atom_info.get_building_block_id()
                for atom_info in atom_infos
            ]

            if build_block_ids[1] is None:
                continue
            # Check if two central atoms of torsion are from the same
            # building block. If not, leave this torsion alone.
            if (build_block_ids[1] != build_block_ids[2]):
                continue

            build_block_torsions = TorsionCalculator().get_results(
                atom_infos[1].get_building_block()
            ).get_torsions()
            atom_map = atom_maps[build_block_ids[1]]

            # Look for a torsion in the building block that has the
            # same central atoms.
            for bb_torsion in build_block_torsions:
                try:
                    matched_atoms = [
                        atom_map[atom_id]
                        for atom_id in bb_torsion.get_atom_ids()
                    ]
                except KeyError:
                    # The atoms of the building block torsion do not
                    # all have corresponding atoms in the constructed
                    # molecule (e.g. atom was deleted in construction)
                    # so skip to next building block torsion.
                    continue
                matched_atom_ids = [
                    atom.get_id()
                    for atom in matched_atoms
                ]
                if set(matched_atom_ids[1:3]) == set(atom_ids[1:3]):
                    # Set the constructed molecule torsion to match the
                    # building block torsion.
                    torsions[i] = Torsion(*matched_atoms)
                    break

        yield tuple(torsions)



import stko
from ..utilities import (
    inequivalent_position_matrices,
    is_equivalent_molecule,
)
import numpy as np


def test_open_babel(case_molecule):
    energy = (
        stko.OpenBabelEnergy('uff').get_energy(case_molecule.molecule)
    )
    optimizer = stko.OpenBabel('uff')
    opt_res = optimizer.optimize(case_molecule.molecule)
    is_equivalent_molecule(opt_res, case_molecule.molecule)
    inequivalent_position_matrices(opt_res, case_molecule.molecule)
    opt_energy = (
        stko.OpenBabelEnergy('uff').get_energy(opt_res)
    )
    assert np.isclose(
        energy, case_molecule.unoptimised_energy, atol=1E-3
    )
    assert opt_energy < case_molecule.unoptimised_energy



from .optimizers import *  # noqa
from .rdkit import *  # noqa
from .gulp import *  # noqa
from .collapser import *  # noqa
from .xtb import *  # noqa
from .macromodel import *  # noqa
from .open_babel import *  # noqa



"""
Utilities for periodic molecules.

"""

import logging
import numpy as np

logger = logging.getLogger(__name__)


def get_approximate_cell_size(molecule, vector_1, vector_2, vector_3):
    """
    Cell size determined from projection of atoms on cell vectors.

    Arguments
    ---------
    molecule : :class:`stk.Molecule`
        Molecule to get approximate cell size of.

    vector_1 : :class:`numpy.ndarray`
        First cell lattice vector of shape (3, ) in Angstrom.

    vector_2 : :class:`numpy.ndarray`
        Second cell lattice vector of shape (3, ) in Angstrom.

    vector_3 : :class:`numpy.ndarray`
        Third cell lattice vector of shape (3, ) in Angstrom.

    Returns
    -------
    max_extent : :class:`float`
        Get maximum multiplier of cell vectors required to fit
        molecule.

    """

    def get_projection(start, target):
        """
        Get the projection of `start` onto `target`.

        """

        return target * np.dot(
            start,
            target,
        ) / np.dot(target, target)

    def get_magnitude_along_vector(atom_pos, vector):
        """
        Get magnitude of atom position along vector.

        """

        projection = get_projection(
            start=atom_pos,
            target=vector
        )

        return np.linalg.norm(np.divide(
            projection,
            vector,
            out=np.zeros_like(projection),
            where=vector != 0
        ))

    # Get projection of all atom positions on three vectors.
    extent_of_vector_1 = 0
    extent_of_vector_2 = 0
    extent_of_vector_3 = 0
    for atom_pos in molecule.get_position_matrix():
        v1_mag = get_magnitude_along_vector(
            atom_pos=atom_pos,
            vector=vector_1
        )
        v2_mag = get_magnitude_along_vector(
            atom_pos=atom_pos,
            vector=vector_2
        )
        v3_mag = get_magnitude_along_vector(
            atom_pos=atom_pos,
            vector=vector_3
        )
        extent_of_vector_1 = max([v1_mag, extent_of_vector_1])
        extent_of_vector_2 = max([v2_mag, extent_of_vector_2])
        extent_of_vector_3 = max([v3_mag, extent_of_vector_3])

    max_extent = max([extent_of_vector_1, extent_of_vector_2])*1.35

    return max_extent


def get_from_parameters(a, b, c, alpha, beta, gamma):
    """
    Create a Lattice using unit cell lengths and angles (in degrees).

    This code is modified from the pymatgen source code [1]_.

    Parameters
    ----------
    a : :class:`float`:
        *a* lattice parameter.

    b : :class:`float`:
        *b* lattice parameter.

    c : :class:`float`:
        *c* lattice parameter.

    alpha : :class:`float`:
        *alpha* angle in degrees.

    beta : :class:`float`:
        *beta* angle in degrees.

    gamma : :class:`float`:
        *gamma* angle in degrees.


    Returns
    -------
    :class:`tuple` of three :class:`numpy.ndarray`
        Tuple of cell lattice vectors of shape (3, ) in Angstrom.

    """

    angles_r = np.radians([alpha, beta, gamma])
    cos_alpha, cos_beta, cos_gamma = np.cos(angles_r)
    sin_alpha, sin_beta, sin_gamma = np.sin(angles_r)

    val = (cos_alpha * cos_beta - cos_gamma) / (sin_alpha * sin_beta)
    # Sometimes rounding errors result in values slightly > 1.
    val = cap_absolute_value(val)
    gamma_star = np.arccos(val)

    vector_a = np.array([a * sin_beta, 0.0, a * cos_beta])
    vector_b = np.array([
        -b * sin_alpha * np.cos(gamma_star),
        b * sin_alpha * np.sin(gamma_star),
        b * cos_alpha,
    ])
    vector_c = np.array([0.0, 0.0, float(c)])

    return tuple([vector_a, vector_b, vector_c])


def cap_absolute_value(value, max_absolute_value=1):
    """
    Returns `value` with absolute value capped at `max_absolute_value`.

    Particularly useful in passing values to trignometric functions
    where numerical errors may result in an argument > 1 being passed
    in.

    This code is modified from the pymatgen source code [1]_.

    Parameters
    ----------
    value : :class:`float`
        Value to cap.

    max_absolute_value : :class:`float`, optional
        Absolute value to cap `value` at.
        Defaults to 1.

    Returns
    -------
    :class:`float`
        `value` capped at `max_absolute_value` with sign preserved.

    References
    ----------
    .. [1] https://pymatgen.org/pymatgen.util.num.html

    """

    return max(min(value, max_absolute_value), -max_absolute_value)



import pytest
import sys
from stko import MacroModelMD, MacroModelForceField


macromodel_check = pytest.mark.skipif(
    all('macromodel_path' not in x for x in sys.argv),
    reason="Only run when MacroModel path is explicitly given."
)


@pytest.fixture(scope="function")
def macromodel_md(macromodel_path):
    return MacroModelMD(macromodel_path)


@pytest.fixture(scope="function")
def macromodel_forcefield(tmpdir, macromodel_path):
    return MacroModelForceField(
        force_field=16,
        output_dir=tmpdir,
        macromodel_path=macromodel_path
    )


@macromodel_check
class TestMacroModelFF:
    def test_generate_com(
        self,
        macromodel_forcefield,
        tmp_path,
        benzene_build
    ):
        # Write the ``.com`` file.
        run_name = 'test_com'
        macromodel_forcefield._generate_com(
            benzene_build, str(tmp_path/run_name)
        )
        with open(str(tmp_path / (run_name+'.com')), 'r') as f:
            com = f.readlines()
        com_path = (
            '/home/sbennett/PhD/stko/tests/optimizers/'
            'macromodel_optimizers/benzene_test.com'
        )
        with open(com_path, 'r') as f:
            test_com = f.readlines()[2:]

        # Do not compare first two lines, as these correspond
        assert test_com == com[2:]



"""
Shape Calculators
=================

#. :class:`.ShapeCalculator`

Wrappers for shape calculators within the :mod:`rdkit` code.

"""

import logging
from rdkit.Chem import Descriptors3D as D3D

from .calculators import Calculator
from .results import ShapeResults


logger = logging.getLogger(__name__)


class ShapeCalculator(Calculator):
    """
    Calculates shape measures of a molecule.

    Uses :mod:`rdkit` 3D Descriptors [1]_ module to calculate all
    measures.

    Examples
    --------
    .. code-block:: python

        import stk
        import stko

        bb1 = stk.BuildingBlock('C1CCCCC1')
        shape_calc = stko.ShapeCalculator()
        shape_results = shape_calc.get_results(bb1)
        eccentricity  = shape_results.get_eccentricity()

    References
    ----------
    .. [1]
    https://www.rdkit.org/docs/source/rdkit.Chem.Descriptors3D.html

    """

    def calculate(self, mol):
        rdkit_mol = mol.to_rdkit_mol()
        results_dict = {
            'pmi1': D3D.PMI1(rdkit_mol),
            'pmi2': D3D.PMI2(rdkit_mol),
            'pmi3': D3D.PMI3(rdkit_mol),
            'npr1': D3D.NPR1(rdkit_mol),
            'npr2': D3D.NPR2(rdkit_mol),
            'asphericity': D3D.Asphericity(rdkit_mol),
            'eccentricity': D3D.Eccentricity(rdkit_mol),
            'inertialshapefactor': D3D.InertialShapeFactor(rdkit_mol),
            'radiusofgyration': D3D.RadiusOfGyration(rdkit_mol),
            'spherocityindex': D3D.SpherocityIndex(rdkit_mol),
        }

        yield results_dict

    def get_results(self, mol):
        """
        Calculate the shape measures of `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The :class:`.Molecule` whose energy is to be calculated.

        Returns
        -------
        :class:`.ShapeResults`
            The shape measures of the molecule.

        """

        return ShapeResults(self.calculate(mol))



"""
RDKit Calculators
=================

#. :class:`.MMFFEnergy`
#. :class:`.UFFEnergy`

Wrappers for calculators within the :mod:`rdkit` code.

"""

import logging
from rdkit.Chem import AllChem as rdkit

from .calculators import Calculator
from .results import EnergyResults


logger = logging.getLogger(__name__)


class MMFFEnergy(Calculator):
    """
    Uses the MMFF force field to calculate energies.

    Examples
    --------
    .. code-block:: python

        import stk
        import stko


        # Create a molecule whose energy we want to know.
        mol1 = stk.BuildingBlock('CCCNCCCN')

        # Create the energy calculator.
        mmff = stko.MMFFEnergy()

        # Calculate the energy.
        results = mmff.get_results(mol1)
        energy = results.get_energy()
        unit_string = results.get_unit_string()

    """

    def __init__(self, ignore_inter_interactions=True):

        self._ignore_inter_interactions = (
            ignore_inter_interactions
        )

    def calculate(self, mol):
        rdkit_mol = mol.to_rdkit_mol()
        rdkit.SanitizeMol(rdkit_mol)
        rdkit.GetSSSR(rdkit_mol)
        ff = rdkit.MMFFGetMoleculeForceField(
            rdkit_mol,
            rdkit.MMFFGetMoleculeProperties(rdkit_mol),
            ignoreInterfragInteractions=self._ignore_inter_interactions
        )
        yield ff.CalcEnergy()

    def get_results(self, mol):
        """
        Calculate the energy of `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The :class:`.Molecule` whose energy is to be calculated.

        Returns
        -------
        :class:`.EnergyResults`
            The energy and units of the energy.

        """

        return EnergyResults(
            generator=self.calculate(mol),
            unit_string='kcal mol-1',
        )

    def get_energy(self, mol):
        """
        Calculate the energy of `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The :class:`.Molecule` whose energy is to be calculated.

        Returns
        -------
        :class:`float`
            The energy.

        """

        return self.get_results(mol).get_energy()


class UFFEnergy(Calculator):
    """
    Uses the UFF force field to calculate energies.

    Examples
    --------
    .. code-block:: python

        import stk
        import stko


        # Create a molecule whose energy we want to know.
        mol1 = stk.BuildingBlock('CCCNCCCN')

        # Create the energy calculator.
        uff = stko.UFFEnergy()

        # Calculate the energy.
        results = uff.get_results(mol1)
        energy = results.get_energy()
        unit_string = results.get_unit_string()

    """

    def __init__(self, ignore_inter_interactions=True):

        self._ignore_inter_interactions = (
            ignore_inter_interactions
        )

    def calculate(self, mol):
        rdkit_mol = mol.to_rdkit_mol()
        rdkit.SanitizeMol(rdkit_mol)
        # RingInfo needs to be initialized, else rdkit may raise an
        # error.
        rdkit.GetSSSR(rdkit_mol)
        ff = rdkit.UFFGetMoleculeForceField(
            mol=rdkit_mol,
            ignoreInterfragInteractions=self._ignore_inter_interactions
        )
        yield ff.CalcEnergy()

    def get_results(self, mol):
        """
        Calculate the energy of `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The :class:`.Molecule` whose energy is to be calculated.

        Returns
        -------
        :class:`.EnergyResults`
            The energy and units of the energy.

        """

        return EnergyResults(
            generator=self.calculate(mol),
            unit_string='kcal mol-1',
        )

    def get_energy(self, mol):
        """
        Calculate the energy of `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The :class:`.Molecule` whose energy is to be calculated.

        Returns
        -------
        :class:`float`
            The energy.

        """

        return self.get_results(mol).get_energy()



"""
XTB Results
===========

#. :class:`.XTBResults`

Results class for the output of XTB.

"""

from .results import Results
from ..extractors import XTBExtractor


class XTBResults(Results):
    """
    Results class containing molecule xTB properties.

    """

    def __init__(self, generator, output_file, extractor=XTBExtractor):

        # Run calculation.
        next(generator)
        self._extractor = extractor(output_file=output_file)

    def get_total_energy(self):
        return (self._extractor.total_energy, 'a.u.')

    def get_homo_lumo_gap(self):
        return (self._extractor.homo_lumo_gap, 'eV')

    def get_fermi_level(self):
        return (self._extractor.fermi_level, 'eV')

    def get_homo_lumo_orbitals(self):
        return (self._extractor.homo_lumo_occ, 'eV')

    def get_qonly_dipole_moments(self):
        return (self._extractor.qonly_dipole_moment, 'Debye')

    def get_full_dipole_moments(self):
        return (self._extractor.full_dipole_moment, 'Debye')

    def get_qonly_quadrupole_moments(self):
        return (self._extractor.qonly_quadrupole_moment, 'Debye')

    def get_qdip_quadrupole_moments(self):
        return (self._extractor.qdip_quadrupole_moment, 'Debye')

    def get_full_quadrupole_moments(self):
        return (self._extractor.full_quadrupole_moment, 'Debye')

    def get_total_free_energy(self):
        try:
            return (self._extractor.total_free_energy, 'a.u.')
        except AttributeError:
            raise AttributeError(
                'Frequency, hessian and thermo calculations not '
                'performed to extract this property.'
            )

    def get_frequencies(self):
        try:
            return (self._extractor.frequencies, 'wavenumber')
        except AttributeError:
            raise AttributeError(
                'Frequency, hessian and thermo calculations not '
                'performed to extract this property.'
            )

    def get_ionisation_potential(self):
        return (self._extractor.ionisation_potential, 'eV')

    def get_electron_affinity(self):
        return (self._extractor.electron_affinity, 'eV')



"""
Dummy Atom
==========

#. :class:`.Du`

A class representing a dummy stk.Atom.

"""

import logging
import stk

logger = logging.getLogger(__name__)


class Du:
    """
    Dummy of an stk.Atom.

    """

    def __init__(self, id):
        """
        Initialize a :class:`.Du`

        Parameters
        ----------

        """

        self._stk_atom = stk.Atom(
            id=id,
            atomic_number=1,
            charge=0,
        )
        self._atomic_number = 0
        self._charge = 0

    def get_id(self):
        """
        Get the ID of the atom.

        Returns
        ------
        :class:`int`
            The atom ID.

        """

        return self._stk_atom.get_id()

    def get_atomic_number(self):
        """
        Get the atomic number of the atom.

        Returns
        ------
        :class:`int`
            The atomic number.

        """

        return self._atomic_number

    def get_charge(self):
        """
        Get the charge of the atom.

        Returns
        ------
        :class:`int`
            The atom charge.

        """

        return self._charge

    def clone(self):
        """
        Returns a clone.

        Returns
        ------
        :class:`.Atom`
            A clone.

        """

        return type(self)(self._stk_atom.get_id())

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self._stk_atom.get_id()})'

    def __str__(self) -> str:
        return repr(self)



def test_get_position(case_data):
    """
    Test :meth:`.PositionedAtom.get_position`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case, containing the atom to test and its correct atomic
        number.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_get_position(case_data.atom, case_data.position)


def _test_get_position(atom, position):
    """
    Test :meth:`.PositionedAtom.get_position`

    Parameters
    ----------
    atom : :class:`.Atom`
        The atom to test.

    position : :class:`tuple` of :class:`float`
        The position (`x`, `y`, `z`) of the atom in cartesian
        coordinates.

    Returns
    -------
    None : :class:`NoneType`

    """

    assert atom.get_position() == position



"""
Positioned Atom
===============

#. :class:`.PositionedAtom`

A class representing a positioned stk.Atom.

"""

import logging

logger = logging.getLogger(__name__)


class PositionedAtom:
    """
    A container for stk.Atom and a coordinate.

    """

    def __init__(self, atom, position):
        """
        Initialize a :class:`PositionedAtom`.

        Parameters
        ----------
        atom : :class:`stk.Atom`
            The atom.

        position : :class:`tuple` of :class:`float`
            The position (`x`, `y`, `z`) of the atom in cartesian
            coordinates.

        """

        self._atom = atom
        self._position = position

    def get_atom(self):
        return self._atom

    def get_atomic_number(self):
        return self._atom.get_atomic_number()

    def get_charge(self):
        return self._atom.get_charge()

    def get_id(self):
        return self._atom.get_id()

    def get_position(self):
        return self._position

    def _with_id(self, id):
        """
        Modify the atom id.

        """

        self._atom = self._atom.with_id(id)
        return self

    def with_id(self, id):
        """
        Get a clone but with a different id.

        Returns
        -------
        :class:`.Atom`
            A clone with a new id. Has the same type as the original
            atom.

        """

        return self.clone()._with_id(id)

    def clone(self):
        """
        Return a clone.

        Returns
        -------
        :class:`.Atom`
            The clone. It has the same type as the original atom.

        """

        clone = self.__class__.__new__(self.__class__)
        clone._atom = self._atom
        clone._position = self._position
        return clone

    def __repr__(self):
        charge = (
            f', charge={self._atom.get_charge()}'
            if self._atom.get_charge() != 0 else ''
        )
        return (
            f'{self._atom.__class__.__name__}({self._atom.get_id()}'
            f'{charge})'
        )

    def __str__(self):
        return repr(self)



"""
XTB Optimizers
==============

#. :class:`.XTB`
#. :class:`.XTBCREST`
#. :class:`.XTBFF`
#. :class:`.XTBFFCREST`

Wrappers for optimizers within the :mod:`xtb` code.

"""

import logging
import os
import shutil
import uuid
import subprocess as sp

from .optimizers import Optimizer
from ..utilities import is_valid_xtb_solvent, XTBInvalidSolventError
from ..calculators.extractors import XTBExtractor


logger = logging.getLogger(__name__)


class XTBOptimizerError(Exception):
    ...


class XTBConvergenceError(XTBOptimizerError):
    ...


class CRESTOptimizerError(Exception):
    ...


class CRESTNotStartedError(CRESTOptimizerError):
    ...


class CRESTNotCompletedError(CRESTOptimizerError):
    ...


class CRESTSettingConflictError(CRESTOptimizerError):
    ...


class XTB(Optimizer):
    """
    Uses GFN-xTB [1]_ to optimize molecules.

    Notes
    -----
    When running :meth:`optimize`, this calculator changes the
    present working directory with :func:`os.chdir`. The original
    working directory will be restored even if an error is raised, so
    unless multi-threading is being used this implementation detail
    should not matter.

    Furthermore, :meth:`optimize` will check that the
    structure is adequately optimized by checking for negative
    frequencies after a Hessian calculation. `max_runs` can be
    provided to the initializer to set the maximum number of
    optimizations which will be attempted at the given
    `opt_level` to obtain an optimized structure. However, we outline
    in the examples how to iterate over `opt_levels` to increase
    convergence criteria and hopefully obtain an optimized structure.
    The presence of negative frequencies can occur even when the
    optimization has converged based on the given `opt_level`.

    *Contributors*
    We thank Andrew Tarzia and Alejandro Santana-Bonilla for their
    contributions to this code.

    Attributes
    ----------
    incomplete : :class:`set` of :class:`.Molecule`
        A :class:`set` of molecules passed to :meth:`optimize` whose
        optimzation was incomplete.

    Examples
    --------
    Note that for :class:`.ConstructedMolecule` objects constructed by
    ``stk``, :class:`XTB` should usually be used in a
    :class:`.OptimizerSequence`. This is because xTB only uses
    xyz coordinates as input and so will not recognize the long bonds
    created during construction. An optimizer which can minimize
    these bonds should be used before :class:`XTB`.

    .. code-block:: python

        import stk
        import stko

        bb1 = stk.BuildingBlock('NCCNCCN', [stk.PrimaryAminoFactory()])
        bb2 = stk.BuildingBlock('O=CCCC=O', [stk.AldehydeFactory()])
        polymer = stk.ConstructedMolecule(
            stk.polymer.Linear(
                building_blocks=(bb1, bb2),
                repeating_unit="AB",
                orientations=[0, 0],
                num_repeating_units=1
            )
        )

        xtb = stko.OptimizerSequence(
            stko.UFF(),
            stko.XTB(xtb_path='/opt/gfnxtb/xtb', unlimited_memory=True)
        )
        polymer = xtb.optimize(polymer)


    By default, all optimizations with xTB are performed using the
    ``--ohess`` flag, which forces the calculation of a numerical
    Hessian, thermodynamic properties and vibrational frequencies.
    :meth:`optimize` will check that the structure is appropriately
    optimized (i.e. convergence is obtained and no negative vibrational
    frequencies are present) and continue optimizing a structure (up to
    `max_runs` times) until this is achieved. This loop, by
    default, will be performed at the same `opt_level`. The
    following example shows how a user may optimize structures with
    tigher convergence criteria (i.e. different `opt_level`)
    until the structure is sufficiently optimized. Furthermore, the
    calculation of the Hessian can be turned off using
    `max_runs` to ``1`` and `calculate_hessian` to ``False``.

    .. code-block:: python

        # Use crude optimization with max_runs=1 because this will
        # not achieve optimization and rerunning it is unproductive.
        xtb_crude = stko.XTB(
            xtb_path='/opt/gfnxtb/xtb',
            output_dir='xtb_crude',
            unlimited_memory=True,
            opt_level='crude',
            max_runs=1,
            calculate_hessian=True
        )
        # Use normal optimization with max_runs == 2.
        xtb_normal = stko.XTB(
            xtb_path='/opt/gfnxtb/xtb',
            output_dir='xtb_normal',
            unlimited_memory=True,
            opt_level='normal',
            max_runs=2
        )
        # Use vtight optimization with max_runs == 2, which should
        # achieve sufficient optimization.
        xtb_vtight = stko.XTB(
            xtb_path='/opt/gfnxtb/xtb',
            output_dir='xtb_vtight',
            unlimited_memory=True,
            opt_level='vtight',
            max_runs=2
        )

        optimizers = [xtb_crude, xtb_normal, xtb_vtight]
        for optimizer in optimizers:
            polymer = optimizer.optimize(polymer)
            if polymer not in optimizer.incomplete:
                break

    References
    ----------
    .. [1] https://xtb-docs.readthedocs.io/en/latest/setup.html

    """

    def __init__(
        self,
        xtb_path,
        gfn_version=2,
        output_dir=None,
        opt_level='normal',
        max_runs=2,
        calculate_hessian=True,
        num_cores=1,
        electronic_temperature=300,
        solvent_model='gbsa',
        solvent=None,
        solvent_grid='normal',
        charge=0,
        num_unpaired_electrons=0,
        unlimited_memory=False,
    ):
        """
        Initialize a :class:`XTB` instance.

        Parameters
        ----------
        xtb_path : :class:`str`
            The path to the xTB executable.

        gfn_version : :class:`int`, optional
            Parameterization of GFN to use in xTB.
            For details see
            https://xtb-docs.readthedocs.io/en/latest/basics.html.

        output_dir : :class:`str`, optional
            The name of the directory into which files generated during
            the optimization are written, if ``None`` then
            :func:`uuid.uuid4` is used.

        opt_level : :class:`str`, optional
            Optimization level to use.
            Can be one of ``'crude'``, ``'sloppy'``, ``'loose'``,
            ``'lax'``, ``'normal'``, ``'tight'``, ``'vtight'``
            or ``'extreme'``.
            For details see
            https://xtb-docs.readthedocs.io/en/latest/optimization.html
            .

        max_runs : :class:`int`, optional
            Maximum number of optimizations to attempt in a row.

        calculate_hessian : :class:`bool`, optional
            Toggle calculation of the hessian and vibrational
            frequencies after optimization. ``True`` is required to
            check that the structure is completely optimized.
            ``False`` will drastically speed up the calculation but
            potentially provide incomplete optimizations and forces
            :attr:`max_runs` to be ``1``.

        num_cores : :class:`int`, optional
            The number of cores xTB should use.

        electronic_temperature : :class:`int`, optional
            Electronic temperature in Kelvin.

        solvent_model : :class:`str`
            Solvent model to use out of older `gbsa` and newer `alpb`.
            `gbsa` is default for backwards compatability, but `alpb`
            is recommended.
            For details see
            https://xtb-docs.readthedocs.io/en/latest/gbsa.html.

        solvent : :class:`str`, optional
            Solvent to use in GBSA implicit solvation method.
            For details see
            https://xtb-docs.readthedocs.io/en/latest/gbsa.html.

        solvent_grid : :class:`str`, optional
            Grid level to use in SASA calculations for GBSA implicit
            solvent.
            Can be one of ``'normal'``, ``'tight'``, ``'verytight'``
            or ``'extreme'``.
            For details see
            https://xtb-docs.readthedocs.io/en/latest/gbsa.html.

        charge : :class:`int`, optional
            Formal molecular charge.

        num_unpaired_electrons : :class:`int`, optional
            Number of unpaired electrons.

        unlimited_memory : :class: `bool`, optional
            If ``True`` :meth:`optimize` will be run without
            constraints on the stack size. If memory issues are
            encountered, this should be ``True``, however this may
            raise issues on clusters.

        """

        if solvent is not None:
            solvent = solvent.lower()
            if gfn_version == 0:
                raise XTBInvalidSolventError(
                    'No solvent valid for version',
                    f' {gfn_version!r}.'
                )
            if not is_valid_xtb_solvent(
                gfn_version=gfn_version,
                solvent_model=solvent_model,
                solvent=solvent,
            ):
                raise XTBInvalidSolventError(
                    f'Solvent {solvent!r} and model {solvent_model!r}',
                    f' is invalid for version {gfn_version!r}.'
                )

        if not calculate_hessian and max_runs != 1:
            max_runs = 1
            logger.warning(
                'Requested that hessian calculation was skipped '
                'but the number of optimizations requested was '
                'greater than 1. The number of optimizations has been '
                'set to 1.'
            )

        self._xtb_path = xtb_path
        self._gfn_version = str(gfn_version)
        self._output_dir = output_dir
        self._opt_level = opt_level
        self._max_runs = max_runs
        self._calculate_hessian = calculate_hessian
        self._num_cores = str(num_cores)
        self._electronic_temperature = str(electronic_temperature)
        self._solvent = solvent
        self._solvent_model = solvent_model
        self._solvent_grid = solvent_grid
        self._charge = str(charge)
        self._num_unpaired_electrons = str(num_unpaired_electrons)
        self._unlimited_memory = unlimited_memory
        self.incomplete = set()

    def _has_neg_frequencies(self, output_file):
        """
        Check for negative frequencies.

        Parameters
        ----------
        output_file : :class:`str`
            Name of output file with xTB results.

        Returns
        -------
        :class:`bool`
            Returns ``True`` if a negative frequency is present.

        """
        xtbext = XTBExtractor(output_file=output_file)
        # Check for one negative frequency, excluding the first
        # 6 frequencies.
        return any(x < 0 for x in xtbext.frequencies[6:])

    def _is_complete(self, output_file):
        """
        Check if xTB optimization has completed and converged.

        Parameters
        ----------
        output_file : :class:`str`
            Name of xTB output file.

        Returns
        -------
        :class:`bool`
            Returns ``False`` if a negative frequency is present.

        Raises
        -------
        :class:`XTBOptimizerError`
            If the optimization failed.

        :class:`XTBConvergenceError`
            If the optimization did not converge.

        """
        if not os.path.exists(output_file):
            # No simulation has been run.
            raise XTBOptimizerError('Optimization failed to start')

        # If convergence is achieved, then .xtboptok should exist.
        if os.path.exists('.xtboptok'):
            # Check for negative frequencies in output file if the
            # hessian was calculated.
            # Return True if there exists at least one.
            if self._calculate_hessian:
                return not self._has_neg_frequencies(output_file)
            else:
                return True

        elif os.path.exists('NOT_CONVERGED'):
            raise XTBConvergenceError('Optimization not converged.')
        else:
            raise XTBOptimizerError('Optimization failed to complete')

    def _run_xtb(self, xyz, out_file):
        """
        Run GFN-xTB.

        Parameters
        ----------
        xyz : :class:`str`
            The name of the input structure ``.xyz`` file.

        out_file : :class:`str`
            The name of output file with xTB results.

        Returns
        -------
        None : :class:`NoneType`

        """

        # Modify the memory limit.
        if self._unlimited_memory:
            memory = 'ulimit -s unlimited ;'
        else:
            memory = ''

        # Set optimization level and type.
        if self._calculate_hessian:
            # Do optimization and check hessian.
            optimization = f'--ohess {self._opt_level}'
        else:
            # Do optimization.
            optimization = f'--opt {self._opt_level}'

        if self._solvent is not None:
            solvent = f'--{self._solvent_model} {self._solvent} '
        else:
            solvent = ''

        cmd = (
            f'{memory} {self._xtb_path} {xyz} '
            f'--gfn {self._gfn_version} '
            f'{optimization} --parallel {self._num_cores} '
            f'--etemp {self._electronic_temperature} '
            f'{solvent} --chrg {self._charge} '
            f'--uhf {self._num_unpaired_electrons} -I det_control.in'
        )

        with open(out_file, 'w') as f:
            # Note that sp.call will hold the program until completion
            # of the calculation.
            sp.call(
                cmd,
                stdin=sp.PIPE,
                stdout=f,
                stderr=sp.PIPE,
                # Shell is required to run complex arguments.
                shell=True
            )

    def _write_detailed_control(self):
        string = f'$gbsa\n   gbsagrid={self._solvent_grid}'

        with open('det_control.in', 'w') as f:
            f.write(string)

    def _run_optimizations(self, mol):
        """
        Run loop of optimizations on `mol` using xTB.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        Returns
        -------
        mol : :class:`.Molecule`
            The optimized molecule.

        opt_complete : :class:`bool`
            Returns ``True`` if the calculation is complete and
            ``False`` if the calculation is incomplete.

        """

        for run in range(self._max_runs):
            xyz = f'input_structure_{run+1}.xyz'
            out_file = f'optimization_{run+1}.output'
            mol.write(xyz)
            self._write_detailed_control()
            self._run_xtb(xyz=xyz, out_file=out_file)
            # Check if the optimization is complete.
            coord_file = 'xtbhess.coord'
            output_xyz = 'xtbopt.xyz'
            opt_complete = self._is_complete(out_file)
            if not opt_complete:
                if os.path.exists(coord_file):
                    # The calculation is incomplete.
                    # Update mol from xtbhess.coord and continue.
                    mol = mol.with_structure_from_file(coord_file)
                else:
                    # Update mol from xtbopt.xyz.
                    mol = mol.with_structure_from_file(output_xyz)
                    # If the negative frequencies are small, then GFN
                    # may not produce the restart file. If that is the
                    # case, exit optimization loop and warn.
                    self.incomplete.add(mol)
                    logging.warning(
                        f'Small negative frequencies present in {mol}.'
                    )
                    return mol, opt_complete
            else:
                # Optimization is complete.
                # Update mol from xtbopt.xyz.
                mol = mol.with_structure_from_file(output_xyz)
                break

        return mol, opt_complete

    def optimize(self, mol):
        """
        Optimize `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        Returns
        -------
        mol : :class:`.Molecule`
            The optimized molecule.

        """

        # Remove mol from self.incomplete if present.
        if mol in self.incomplete:
            self.incomplete.remove(mol)

        if self._output_dir is None:
            output_dir = str(uuid.uuid4().int)
        else:
            output_dir = self._output_dir
        output_dir = os.path.abspath(output_dir)

        if os.path.exists(output_dir):
            shutil.rmtree(output_dir)

        os.mkdir(output_dir)
        init_dir = os.getcwd()
        os.chdir(output_dir)

        try:
            mol, complete = self._run_optimizations(mol)
        finally:
            os.chdir(init_dir)

        if not complete:
            self.incomplete.add(mol)
            logging.warning(f'Optimization is incomplete for {mol}.')

        return mol


class XTBCREST(Optimizer):
    """
    Uses GFN-n [1]_ to run CREST [2]_ on molecules.

    Notes
    -----
    Requires version > 6.2 of xtb.

    When running :meth:`optimize`, this calculator changes the
    present working directory with :func:`os.chdir`. The original
    working directory will be restored even if an error is raised, so
    unless multi-threading is being used this implementation detail
    should not matter.

    *Restrictions to iMTD-GC Algorithm*
    Z-matrix sorting is forced to be off because :class:`stk.Molecules`
    cannot have their atom ordering changed by an external program at
    this stage.

    Examples
    --------

    Note that for :class:`.ConstructedMolecule` objects constructed by
    ``stk``, :class:`XTBCREST` should usually be used in a
    :class:`.OptimizerSequence`. This is because xTB only uses
    xyz coordinates as input and so will not recognize the long bonds
    created during construction. An optimizer which can minimize
    these bonds should be used before :class:`XTBCREST`. Further,
    CREST runs best on an input structure optimized at the same level
    used throughout the algorithm (i.e. :class:`XTB`).

    .. code-block:: python

        import stk
        import stko

        bb1 = stk.BuildingBlock('NCCNCCN', [stk.PrimaryAminoFactory()])
        bb2 = stk.BuildingBlock('O=CCCC=O', [stk.AldehydeFactory()])
        polymer = stk.ConstructedMolecule(
            stk.polymer.Linear(
                building_blocks=(bb1, bb2),
                repeating_unit="AB",
                orientations=[0, 0],
                num_repeating_units=1
            )
        )

        xtb = stko.OptimizerSequence(
            stko.UFF(),
            stko.XTB(
                xtb_path='/opt/gfnxtb/xtb',
                unlimited_memory=True,
            ),
            # Perform quick conformer search.
            stko.XTBCREST(
                crest_path='/opt/crest/crest',
                xtb_path='/opt/gfnxtb/xtb',
                opt_level='normal',
                speed_setting='quick',
                unlimited_memory=True,
            ),
        )
        polymer = xtb.optimize(polymer)

    While most auxiliary and MD run files will be deleted if
    :attr:`keepdir` is `False`, the files listed below are kept in
    :attr:`output_dir` and could be useful for further analysis!

        | crest_best.xyz: Structure of the lowest energy conformer.
        |    This structure is output by :meth:`optimize`.
        | crest_conformers.xyz: Structure of all conformers.
        |    All conformers with RMSD and energy threshold of the
        |    lowest energy conformer.
        | crest_rotamers.xyz: Structure of all rotamers.
        |    All unique rotamers explored by CREST.
        | crest.energies: Relative conformer energies in a.u..
        | gfn_topo: GFN-FF binary topology file.
        |    Defines the molecules force field topology.

    References
    ----------
    .. [1] https://xtb-docs.readthedocs.io/en/latest/setup.html
    .. [2] https://xtb-docs.readthedocs.io/en/latest/crestcmd.html

    """

    def __init__(
        self,
        crest_path,
        xtb_path,
        gfn_version=2,
        output_dir=None,
        opt_level='normal',
        md_len=None,
        ewin=5,
        speed_setting=None,
        keepdir=False,
        num_cores=4,
        cross=True,
        charge=0,
        electronic_temperature=300,
        solvent_model='gbsa',
        solvent=None,
        num_unpaired_electrons=0,
        unlimited_memory=False,
    ):
        """
        Initialize a :class:`XTBCREST` instance.

        Parameters
        ----------
        crest_path : :class:`str`
            The path to the CREST executable.

        xtb_path : :class:`str`
            The path to the xTB executable.
            Version >6.3.0 is required.

        gfn_version : :class:`int`, optional
            Parameterization of GFN to use in xTB.
            For details see
            https://xtb-docs.readthedocs.io/en/latest/basics.html.

        output_dir : :class:`str`, optional
            The name of the directory into which files generated during
            the optimization are written, if ``None`` then
            :func:`uuid.uuid4` is used.

        opt_level : :class:`str`, optional
            Optimization level to use.
            Can be one of ``'crude'``, ``'sloppy'``, ``'loose'``,
            ``'lax'``, ``'normal'``, ``'tight'``, ``'vtight'``
            or ``'extreme'``.
            For details see
            https://xtb-docs.readthedocs.io/en/latest/optimization.html
            .

        md_len : :class:`float`, optional.
            Set length of the meta-dynamics simulations (MTD) in ps.
            Default is chosen based on size and flexibility of the
            system.

        ewin : :class:`float`, optional.
            Set the energy threshold in kcal/mol for conformer
            selection. Double this is used in crude optimization.
            Defaults ot 5 kcal/mol and is overridden by
            :attr:`speed_setting`.

        speed_setting : :class:`str` or :class:`NoneType`, optional
            Conformer search speed setting. Fast methods turn off
            parts of the calculations and alter MD run times.
            Defaults to no modification of iMTD-GC algorithm: `None`.
            Can be one of ``'norotmd'``, ``'quick'``, ``'squick'``
            or ``'mquick'``.
            Overrides :attr:`ewin` with chosen parameters.
            For details see
            https://xtb-docs.readthedocs.io/en/latest/crestcmd.html.

        keepdir : :class:`bool`, optional
            `True` to keep subdirectories from MD runs.
            Defaults to `False`.
            For details see
            https://xtb-docs.readthedocs.io/en/latest/crestcmd.html.

        num_cores : :class:`int`, optional
            The number of cores CREST should use.

        charge : :class:`int`, optional
            Formal molecular charge.

        electronic_temperature : :class:`int`, optional
            Electronic temperature in Kelvin.

        solvent_model : :class:`str`
            Solvent model to use out of older `gbsa` and newer `alpb`.
            `gbsa` is default for backwards compatability, but `alpb`
            is recommended.
            For details see
            https://xtb-docs.readthedocs.io/en/latest/gbsa.html.

        solvent : :class:`str`, optional
            Solvent to use in GBSA implicit solvation method.
            For details see
            https://xtb-docs.readthedocs.io/en/latest/gbsa.html.

        num_unpaired_electrons : :class:`int`, optional
            Number of unpaired electrons.

        cross : :class:`bool`, optional
            Whether or not structure crossing is performed.

        unlimited_memory : :class: `bool`, optional
            If ``True`` :meth:`optimize` will be run without
            constraints on the stack size. If memory issues are
            encountered, this should be ``True``, however this may
            raise issues on clusters.

        """

        if solvent is not None:
            solvent = solvent.lower()
            if gfn_version == 0:
                raise XTBInvalidSolventError(
                    'No solvent valid for version',
                    f' {gfn_version!r}.'
                )
            if not is_valid_xtb_solvent(
                gfn_version=gfn_version,
                solvent_model=solvent_model,
                solvent=solvent,
            ):
                raise XTBInvalidSolventError(
                    f'Solvent {solvent!r} and model {solvent_model!r}',
                    f' is invalid for version {gfn_version!r}.'
                )

        self._crest_path = crest_path
        self._xtb_path = xtb_path
        self._gfn_version = str(gfn_version)
        self._output_dir = output_dir
        self._opt_level = opt_level
        self._mdlen = md_len

        if speed_setting is not None and ewin != 5:
            raise CRESTSettingConflictError(
                f'The chosen speed setting {speed_setting} will ',
                f'override the chosen energy window {ewin}.'
            )

        self._ewin = ewin
        self._speed_setting = speed_setting
        self._keepdir = keepdir
        self._num_cores = str(num_cores)
        self._electronic_temperature = str(electronic_temperature)
        self._solvent_model = solvent_model
        self._solvent = solvent
        self._charge = str(charge)
        self._num_unpaired_electrons = str(num_unpaired_electrons)
        self._cross = cross
        self._unlimited_memory = unlimited_memory

    def _is_complete(self, output_file, output_xyzs):
        """
        Check if CREST run has completed.

        Parameters
        ----------
        output_file : :class:`str`
            Name of CREST output file.

        output_xyzs : :class:`str`
            Name of CREST conformer output files.
            crest_best.xyz > Best conformer, exists throughout run.
            crest_conformers.xyz > All conformers,
                exists throughout run.

        Returns
        -------
        :class:`bool`
            Returns ``False`` if a negative frequency is present.

        Raises
        -------
        :class:`CRESTNotStartedError`
            If the CREST run failed to start.

        :class:`CRESTNotCompletedError`
            If the CREST run failed to complete.

        """

        if not os.path.exists(output_file):
            # No simulation has been run.
            raise CRESTNotStartedError('CREST run did not start')

        elif any(not os.path.exists(i) for i in output_xyzs):
            # Best conformer was not output.
            raise CRESTNotCompletedError('CREST run did not complete')

        return True

    def _run_crest(self, xyz, out_file):
        """
        Run CREST along side GFN-xTB.

        Parameters
        ----------
        xyz : :class:`str`
            The name of the input structure ``.xyz`` file.

        out_file : :class:`str`
            The name of output file with xTB results.

        Returns
        -------
        None : :class:`NoneType`

        """

        # Modify the memory limit.
        if self._unlimited_memory:
            memory = 'ulimit -s unlimited ;'
        else:
            memory = ''

        if self._solvent is not None:
            solvent = f'--{self._solvent_model} {self._solvent}'
        else:
            solvent = ''

        # Set optimization level and type.
        optimization = f'-opt {self._opt_level}'
        mdlen = '' if self._mdlen is None else f'-mdlen {self._mdlen} '
        keepdirs = '-keepdir' if self._keepdir is True else ''
        if self._speed_setting is not None:
            speed_settings = f'-{self._speed_setting}'
            ewin = ''
        else:
            speed_settings = ''
            ewin = f'-ewin {self._ewin} '
        cross = ('-nocross' if self._cross is False else '')

        cmd = (
            f'{memory} {self._crest_path} {xyz} '
            f'-xnam {self._xtb_path} -nozs {cross} '
            f'{ewin} {mdlen}'
            f'{solvent} -chrg {self._charge} '
            f'-uhf {self._num_unpaired_electrons} '
            f'-gfn{self._gfn_version} '
            f'{speed_settings} '
            f'{optimization} -T {self._num_cores} {keepdirs}'
        )

        with open(out_file, 'w') as f:
            # Note that sp.call will hold the program until completion
            # of the calculation.
            sp.call(
                cmd,
                stdin=sp.PIPE,
                stdout=f,
                stderr=sp.PIPE,
                # Shell is required to run complex arguments.
                shell=True
            )

    def _run_optimization(self, mol):
        """
        Run loop of optimizations on `mol` using xTB.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        Returns
        -------
        mol : :class:`.Molecule`
            The optimized molecule.

        opt_complete : :class:`bool`
            Returns ``True`` if the calculation is complete and
            ``False`` if the calculation is incomplete.

        """

        xyz = 'input_structure.xyz'
        out_file = 'crest.output'
        mol.write(xyz)
        self._run_crest(xyz=xyz, out_file=out_file)

        # Check if the optimization is complete.
        output_xyzs = [
            'crest_best.xyz',
            'crest_conformers.xyz',
        ]
        opt_complete = self._is_complete(out_file, output_xyzs)
        mol = mol.with_structure_from_file(output_xyzs[0])

        return mol, opt_complete

    def optimize(self, mol):
        """
        Optimize `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        Returns
        -------
        mol : :class:`.Molecule`
            The optimized molecule.

        """

        if self._output_dir is None:
            output_dir = str(uuid.uuid4().int)
        else:
            output_dir = self._output_dir
        output_dir = os.path.abspath(output_dir)

        if os.path.exists(output_dir):
            shutil.rmtree(output_dir)

        os.mkdir(output_dir)
        init_dir = os.getcwd()
        os.chdir(output_dir)

        try:
            mol, complete = self._run_optimization(mol)
        finally:
            os.chdir(init_dir)

        if not complete:
            logging.warning(f'CREST run is incomplete for {mol}.')

        return mol


class XTBFF(Optimizer):
    """
    Uses GFN-FF [1]_ to optimize molecules.

    Notes
    -----
    GFN-FF requires >= version 6.3 of xtb.

    When running :meth:`optimize`, this calculator changes the
    present working directory with :func:`os.chdir`. The original
    working directory will be restored even if an error is raised, so
    unless multi-threading is being used this implementation detail
    should not matter.

    Currently, we only provide inputs that work with GFN-FF,
    specifically the charge of the system. Other electronic properties
    of the molecule are not relavent to a forcefield optimisation.

    Examples
    --------
    Note that for :class:`.ConstructedMolecule` objects constructed by
    ``stk``, :class:`XTBFF` should usually be used in a
    :class:`.OptimizerSequence`. This is because xTB only uses
    xyz coordinates as input and so will not recognize the long bonds
    created during construction. An optimizer which can minimize
    these bonds should be used before :class:`XTBFF`.

    .. code-block:: python

        import stk
        import stko

        bb1 = stk.BuildingBlock('NCCNCCN', [stk.PrimaryAminoFactory()])
        bb2 = stk.BuildingBlock('O=CCCC=O', [stk.AldehydeFactory()])
        polymer = stk.ConstructedMolecule(
            stk.polymer.Linear(
                building_blocks=(bb1, bb2),
                repeating_unit="AB",
                orientations=[0, 0],
                num_repeating_units=1
            )
        )

        xtb = stko.OptimizerSequence(
            stko.UFF(),
            stko.XTBFF(
                xtb_path='/opt/gfnxtb/xtb',
                unlimited_memory=True,
            ),
        )
        polymer = xtb.optimize(polymer)

    References
    ----------
    .. [1] https://xtb-docs.readthedocs.io/en/latest/gfnff.html

    """

    def __init__(
        self,
        xtb_path,
        output_dir=None,
        opt_level='normal',
        num_cores=1,
        charge=0,
        unlimited_memory=False,
    ):
        """
        Initialize a :class:`XTB` instance.

        Parameters
        ----------
        xtb_path : :class:`str`
            The path to the xTB executable.

        output_dir : :class:`str`, optional
            The name of the directory into which files generated during
            the optimization are written, if ``None`` then
            :func:`uuid.uuid4` is used.

        opt_level : :class:`str`, optional
            Optimization level to use.
            Can be one of ``'crude'``, ``'sloppy'``, ``'loose'``,
            ``'lax'``, ``'normal'``, ``'tight'``, ``'vtight'``
            or ``'extreme'``.
            For details see
            https://xtb-docs.readthedocs.io/en/latest/optimization.html
            .

        num_cores : :class:`int`, optional
            The number of cores xTB should use.

        charge : :class:`int`, optional
            Formal molecular charge.

        unlimited_memory : :class: `bool`, optional
            If ``True`` :meth:`optimize` will be run without
            constraints on the stack size. If memory issues are
            encountered, this should be ``True``, however this may
            raise issues on clusters.

        """

        self._xtb_path = xtb_path
        self._output_dir = output_dir
        self._opt_level = opt_level
        self._num_cores = str(num_cores)
        self._charge = str(charge)
        self._unlimited_memory = unlimited_memory

    def _is_complete(self, output_file):
        """
        Check if xTB optimization has completed and converged.

        Parameters
        ----------
        output_file : :class:`str`
            Name of xTB output file.

        Returns
        -------
        :class:`bool`
            Returns ``False`` if a negative frequency is present.

        Raises
        -------
        :class:`XTBOptimizerError`
            If the optimization failed.

        :class:`XTBConvergenceError`
            If the optimization did not converge.

        """
        if not os.path.exists(output_file):
            # No simulation has been run.
            raise XTBOptimizerError('Optimization failed to start')
        # If convergence is achieved, then .xtboptok should exist.
        if os.path.exists('.xtboptok'):
            return True
        elif os.path.exists('NOT_CONVERGED'):
            raise XTBConvergenceError('Optimization not converged.')
        else:
            raise XTBOptimizerError('Optimization failed to complete')

    def _run_xtb(self, xyz, out_file):
        """
        Run GFN-xTB.

        Parameters
        ----------
        xyz : :class:`str`
            The name of the input structure ``.xyz`` file.

        out_file : :class:`str`
            The name of output file with xTB results.

        Returns
        -------
        None : :class:`NoneType`

        """

        # Modify the memory limit.
        if self._unlimited_memory:
            memory = 'ulimit -s unlimited ;'
        else:
            memory = ''

        # Set optimization level and type.
        optimization = f'--opt {self._opt_level}'

        cmd = (
            f'{memory} {self._xtb_path} {xyz} '
            f'--gfnff '
            f'{optimization} --parallel {self._num_cores} '
            f'--chrg {self._charge} '
        )

        with open(out_file, 'w') as f:
            # Note that sp.call will hold the program until completion
            # of the calculation.
            sp.call(
                cmd,
                stdin=sp.PIPE,
                stdout=f,
                stderr=sp.PIPE,
                # Shell is required to run complex arguments.
                shell=True
            )

    def _run_optimization(self, mol):
        """
        Run loop of optimizations on `mol` using xTB.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        Returns
        -------
        mol : :class:`.Molecule`
            The optimized molecule.

        opt_complete : :class:`bool`
            Returns ``True`` if the calculation is complete and
            ``False`` if the calculation is incomplete.

        """

        xyz = 'input_structure_ff.xyz'
        out_file = 'optimization_ff.output'
        mol.write(xyz)
        self._run_xtb(xyz=xyz, out_file=out_file)

        # Check if the optimization is complete.
        output_xyz = 'xtbopt.xyz'
        opt_complete = self._is_complete(out_file)
        mol = mol.with_structure_from_file(output_xyz)

        return mol, opt_complete

    def optimize(self, mol):
        """
        Optimize `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        Returns
        -------
        mol : :class:`.Molecule`
            The optimized molecule.

        """

        if self._output_dir is None:
            output_dir = str(uuid.uuid4().int)
        else:
            output_dir = self._output_dir
        output_dir = os.path.abspath(output_dir)

        if os.path.exists(output_dir):
            shutil.rmtree(output_dir)

        os.mkdir(output_dir)
        init_dir = os.getcwd()
        os.chdir(output_dir)

        try:
            mol, complete = self._run_optimization(mol)
        finally:
            os.chdir(init_dir)

        if not complete:
            logging.warning(f'Optimization is incomplete for {mol}.')

        return mol


class XTBFFCREST(Optimizer):
    """
    Uses GFN-FF [1]_ to run CREST [2]_ on molecules.

    Notes
    -----
    GFN-FF requires version 6.3 of xtb.

    When running :meth:`optimize`, this calculator changes the
    present working directory with :func:`os.chdir`. The original
    working directory will be restored even if an error is raised, so
    unless multi-threading is being used this implementation detail
    should not matter.

    Currently, we only provide inputs that work with GFN-FF,
    specifically the charge of the system. Other electronic properties
    of the molecule are not relavent to a forcefield optimisation.
    We intend on adding more options in the future!

    *Restrictions to iMTD-GC Algorithm*
    Z-matrix sorting is forced to be off because :class:`stk.Molecules`
    cannot have their atom ordering changed by an external program at
    this stage.

    Examples
    --------

    Note that for :class:`.ConstructedMolecule` objects constructed by
    ``stk``, :class:`XTBFFCREST` should usually be used in a
    :class:`.OptimizerSequence`. This is because xTB only uses
    xyz coordinates as input and so will not recognize the long bonds
    created during construction. An optimizer which can minimize
    these bonds should be used before :class:`XTBFFCREST`. Further,
    CREST runs best on an input structure optimized at the same level
    used throughout the algorithm (i.e. :class:`XTBFF`).

    .. code-block:: python

        import stk
        import stko

        bb1 = stk.BuildingBlock('NCCNCCN', [stk.PrimaryAminoFactory()])
        bb2 = stk.BuildingBlock('O=CCCC=O', [stk.AldehydeFactory()])
        polymer = stk.ConstructedMolecule(
            stk.polymer.Linear(
                building_blocks=(bb1, bb2),
                repeating_unit="AB",
                orientations=[0, 0],
                num_repeating_units=1
            )
        )

        xtb = stko.OptimizerSequence(
            stko.UFF(),
            stko.XTBFF(
                xtb_path='/opt/gfnxtb/xtb',
                unlimited_memory=True,
            ),
            # Perform quick conformer search.
            stko.XTBFFCREST(
                crest_path='/opt/crest/crest',
                xtb_path='/opt/gfnxtb/xtb',
                opt_level='normal',
                speed_setting='quick',
                unlimited_memory=True,
            ),
        )
        polymer = xtb.optimize(polymer)

    While most auxiliary and MD run files will be deleted if
    :attr:`keepdir` is `False`, the files listed below are kept in
    :attr:`output_dir` and could be useful for further analysis!

        | crest_best.xyz: Structure of the lowest energy conformer.
        |    This structure is output by :meth:`optimize`.
        | crest_conformers.xyz: Structure of all conformers.
        |    All conformers with RMSD and energy threshold of the
        |    lowest energy conformer.
        | crest_rotamers.xyz: Structure of all rotamers.
        |    All unique rotamers explored by CREST.
        | crest.energies: Relative conformer energies in a.u..
        | gfn_topo: GFN-FF binary topology file.
        |    Defines the molecules force field topology.

    References
    ----------
    .. [1] https://xtb-docs.readthedocs.io/en/latest/gfnff.html
    .. [2] https://xtb-docs.readthedocs.io/en/latest/crestcmd.html

    """

    def __init__(
        self,
        crest_path,
        xtb_path,
        output_dir=None,
        opt_level='normal',
        md_len=None,
        ewin=5,
        speed_setting=None,
        keepdir=False,
        num_cores=1,
        charge=0,
        cross=True,
        unlimited_memory=False,
    ):
        """
        Initialize a :class:`XTBFFCREST` instance.

        Parameters
        ----------
        crest_path : :class:`str`
            The path to the CREST executable.

        xtb_path : :class:`str`
            The path to the xTB executable.
            Version >6.3.0 is required.

        output_dir : :class:`str`, optional
            The name of the directory into which files generated during
            the optimization are written, if ``None`` then
            :func:`uuid.uuid4` is used.

        opt_level : :class:`str`, optional
            Optimization level to use.
            Can be one of ``'crude'``, ``'sloppy'``, ``'loose'``,
            ``'lax'``, ``'normal'``, ``'tight'``, ``'vtight'``
            or ``'extreme'``.
            For details see
            https://xtb-docs.readthedocs.io/en/latest/optimization.html
            .

        md_len : :class:`float`, optional.
            Set length of the meta-dynamics simulations (MTD) in ps.
            Default is chosen based on size and flexibility of the
            system.

        ewin : :class:`float`, optional.
            Set the energy threshold in kcal/mol for conformer
            selection. Double this is used in crude optimization.
            Defaults ot 5 kcal/mol and is overridden by
            :attr:`speed_setting`.

        speed_setting : :class:`str` or :class:`NoneType`, optional
            Conformer search speed setting. Fast methods turn off
            parts of the calculations and alter MD run times.
            Defaults to no modification of iMTD-GC algorithm: `None`.
            Can be one of ``'norotmd'``, ``'quick'``, ``'squick'``
            or ``'mquick'``.
            Overrides :attr:`ewin` with chosen parameters.
            For details see
            https://xtb-docs.readthedocs.io/en/latest/crestcmd.html.

        keepdir : :class:`bool`, optional
            `True` to keep subdirectories from MD runs.
            Defaults to `False`.
            For details see
            https://xtb-docs.readthedocs.io/en/latest/crestcmd.html.

        num_cores : :class:`int`, optional
            The number of cores CREST should use.

        charge : :class:`int`, optional
            Formal molecular charge.

        cross : :class:`bool`, optional
            Whether or not structure crossing is performed.

        unlimited_memory : :class: `bool`, optional
            If ``True`` :meth:`optimize` will be run without
            constraints on the stack size. If memory issues are
            encountered, this should be ``True``, however this may
            raise issues on clusters.

        """

        self._crest_path = crest_path
        self._xtb_path = xtb_path
        self._output_dir = output_dir
        self._opt_level = opt_level
        self._mdlen = md_len

        if speed_setting is not None and ewin != 5:
            raise CRESTSettingConflictError(
                f'The chosen speed setting {speed_setting} will ',
                f'override the chosen energy window {ewin}.'
            )

        self._ewin = ewin
        self._speed_setting = speed_setting
        self._keepdir = keepdir
        self._num_cores = str(num_cores)
        self._charge = str(charge)
        self._cross = cross
        self._unlimited_memory = unlimited_memory

    def _is_complete(self, output_file, output_xyzs):
        """
        Check if CREST run has completed.

        Parameters
        ----------
        output_file : :class:`str`
            Name of CREST output file.

        output_xyzs : :class:`str`
            Name of CREST conformer output files.
            crest_best.xyz > Best conformer, exists throughout run.
            crest_conformers.xyz > All conformers,
                exists throughout run.

        Returns
        -------
        :class:`bool`
            Returns ``False`` if a negative frequency is present.

        Raises
        -------
        :class:`CRESTNotStartedError`
            If the CREST run failed to start.

        :class:`CRESTNotCompletedError`
            If the CREST run failed to complete.

        """

        if not os.path.exists(output_file):
            # No simulation has been run.
            raise CRESTNotStartedError('CREST run did not start')

        elif any(not os.path.exists(i) for i in output_xyzs):
            # Best conformer was not output.
            raise CRESTNotCompletedError('CREST run did not complete')

        return True

    def _run_crest(self, xyz, out_file):
        """
        Run CREST along side GFN-xTB.

        Parameters
        ----------
        xyz : :class:`str`
            The name of the input structure ``.xyz`` file.

        out_file : :class:`str`
            The name of output file with xTB results.

        Returns
        -------
        None : :class:`NoneType`

        """

        # Modify the memory limit.
        if self._unlimited_memory:
            memory = 'ulimit -s unlimited ;'
        else:
            memory = ''

        # Set optimization level and type.
        optimization = f'-opt {self._opt_level}'
        mdlen = '' if self._mdlen is None else f'-mdlen {self._mdlen} '
        keepdirs = '-keepdir' if self._keepdir is True else ''
        if self._speed_setting is not None:
            speed_settings = f'-{self._speed_setting}'
            ewin = ''
        else:
            speed_settings = ''
            ewin = f'-ewin {self._ewin} '
        cross = ('-nocross' if self._cross is False else '')

        cmd = (
            f'{memory} {self._crest_path} {xyz} '
            f'-xnam {self._xtb_path} -nozs {cross} '
            f'{ewin} {mdlen}'
            f'-chrg {self._charge} '
            f'-gff {speed_settings} '
            f'{optimization} -T {self._num_cores} {keepdirs}'
        )

        with open(out_file, 'w') as f:
            # Note that sp.call will hold the program until completion
            # of the calculation.
            sp.call(
                cmd,
                stdin=sp.PIPE,
                stdout=f,
                stderr=sp.PIPE,
                # Shell is required to run complex arguments.
                shell=True
            )

    def _run_optimization(self, mol):
        """
        Run loop of optimizations on `mol` using xTB.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        Returns
        -------
        mol : :class:`.Molecule`
            The optimized molecule.

        opt_complete : :class:`bool`
            Returns ``True`` if the calculation is complete and
            ``False`` if the calculation is incomplete.

        """

        xyz = 'input_structure_ff.xyz'
        out_file = 'crest_ff.output'
        mol.write(xyz)
        self._run_crest(xyz=xyz, out_file=out_file)

        # Check if the optimization is complete.
        output_xyzs = [
            'crest_best.xyz',
            'crest_conformers.xyz',
        ]
        opt_complete = self._is_complete(out_file, output_xyzs)
        mol = mol.with_structure_from_file(output_xyzs[0])

        return mol, opt_complete

    def optimize(self, mol):
        """
        Optimize `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        Returns
        -------
        mol : :class:`.Molecule`
            The optimized molecule.

        """

        if self._output_dir is None:
            output_dir = str(uuid.uuid4().int)
        else:
            output_dir = self._output_dir
        output_dir = os.path.abspath(output_dir)

        if os.path.exists(output_dir):
            shutil.rmtree(output_dir)

        os.mkdir(output_dir)
        init_dir = os.getcwd()
        os.chdir(output_dir)

        try:
            mol, complete = self._run_optimization(mol)
        finally:
            os.chdir(init_dir)

        if not complete:
            logging.warning(f'CREST run is incomplete for {mol}.')

        return mol



"""
GULP Metal Optimizer
====================

#. :class:`.GulpOptimizer`
#. :class:`.GulpMDOpimizer`

Wrappers for calculators within the :mod:`gulp` code.

Examples
--------

While metal atoms are not required, UFF4MOF is useful because it
encompasses almost all chemical environments commonly found in
metal-organic structures. Better forcefields exist for purely
organic molecules! An interface with GULP is provided, which takes
the forcefield types assigned by RDKit for non-metal atoms and
user defined forcefield types for metal atoms to perform geometry
optimisations.

.. code-block:: python

    import stk
    import stko
    from rdkit.Chem import AllChem as rdkit

    # Produce a Pd+2 atom with 4 functional groups.
    atom = rdkit.MolFromSmiles('[Pd+2]')
    atom.AddConformer(rdkit.Conformer(atom.GetNumAtoms()))
    palladium_atom = stk.BuildingBlock.init_from_rdkit_mol(atom)
    atom_0, = palladium_atom.get_atoms(0)
    palladium_atom = palladium_atom.with_functional_groups(
        (stk.SingleAtom(atom_0) for i in range(4))
    )

    # Build a building block with two functional groups using
    # the SmartsFunctionalGroupFactory.
    bb1 = stk.BuildingBlock(
        smiles=('C1=CC(=CC(=C1)C2=CN=CC=C2)C3=CN=CC=C3'),
        functional_groups=[
            stk.SmartsFunctionalGroupFactory(
                smarts='[#6]~[#7X2]~[#6]',
                bonders=(1, ),
                deleters=(),
            ),
        ],
    )

    # Build a metal-organic cage with dative bonds between
    # GenericFunctionalGroup and SingleAtom functional groups.
    cage = stk.ConstructedMolecule(
        stk.cage.M2L4Lantern(
            building_blocks={
                palladium_atom: (0, 1),
                bb1: (2, 3, 4, 5)
            },
            reaction_factory=stk.DativeReactionFactory(
                stk.GenericReactionFactory(
                    bond_orders={
                        frozenset({
                            stk.GenericFunctionalGroup,
                            stk.SingleAtom
                        }): 9
                    }
                )
            )
        )
    )

    # Perform Gulp optimisation with UFF4MOF.
    # Use conjugate gradient method for a slower, but more stable
    # optimisation.

    gulp_opt = stko.GulpUFFOptimizer(
        gulp_path='path/to/gulp',
        metal_FF={46: 'Pd4+2'},
        conjugate_gradient=True
    )

    # Assign the force field.
    gulp_opt.assign_FF(cage)
    # Run optimization.
    cage = gulp_opt.optimize(mol=cage)

Conformer searching is often useful, so we have provided an interface
to MD simulations using GULP and UFF4MOF. A conformer search can be
run at high temperature, where N conformers are extracted at constant
intervals throughtout the simulation and optimized using UFF4MOF. The
lowest energy conformer is returned. After these MD steps, it is
crucial to reoptimize the resultant structure using a better forcefield
or a more robust method!

.. code-block:: python

    gulp_MD = stko.GulpUFFMDOptimizer(
        gulp_path='path/to/gulp',
        metal_FF={46: 'Pd4+2'},
        temperature=300,
        N_conformers=10,
        opt_conformers=True,
    )
    gulp_MD.assign_FF(cage)
    cage = gulp_MD.optimize(cage)

"""

import logging
import re
import uuid
import os
import shutil
import subprocess as sp
from rdkit.Chem import AllChem as rdkit

from .optimizers import Optimizer
from ..utilities import (
    has_h_atom,
    has_metal_atom,
    get_metal_atoms,
    get_metal_bonds,
    to_rdkit_mol_without_metals
)


logger = logging.getLogger(__name__)


class ExpectedMetal(Exception):
    ...


class UFFTyperError(Exception):
    ...


class GulpUFFOptimizer(Optimizer):
    """
    Applies forcefield optimizers that can handle metal centres.

    Notes
    -----
    By default, :meth:`optimize` will run an optimisation using the
    UFF4MOF. This forcefield requires some explicit metal atom
    definitions, which are determined by the user.

    """

    def __init__(
        self,
        gulp_path,
        maxcyc=1000,
        metal_FF=None,
        metal_ligand_bond_order=None,
        conjugate_gradient=False,
        output_dir=None,
    ):
        """
        Initialize a :class:`GulpUFFOptimizer` instance.

        Parameters
        ----------
        gulp_path : :class:`str`
            Path to GULP executable.

        maxcyc : :class:`int`
            Set the maximum number of optimisation steps to use.
            Default in Gulp is 1000.

        metal_FF : :class:`dict`, optional
            Dictionary with metal atom forcefield assignments.
            Key: :class:`int` : atomic number.
            Value: :class:`str` : UFF4MOF forcefield type.

        metal_ligand_bond_order : :class:`str`, optional
            Bond order to use for metal-ligand bonds. Defaults to
            `half`, but using `resonant` can increase the force
            constant for stronger metal-ligand interactions.

        conjugate_gradient : :class:`bool`, optional
            ``True`` to use Conjugate Graditent method.
            Defaults to ``False``

        output_dir : :class:`str`, optional
            The name of the directory into which files generated during
            the calculation are written, if ``None`` then
            :func:`uuid.uuid4` is used.

        """

        self._gulp_path = gulp_path
        self._maxcyc = maxcyc
        self._metal_FF = metal_FF
        self._metal_ligand_bond_order = (
            'half' if metal_ligand_bond_order is None
            else metal_ligand_bond_order
        )
        self._conjugate_gradient = conjugate_gradient
        self._output_dir = output_dir

    def _add_atom_charge_flags(self, atom, atomkey):
        """
        Add atom charge flags for forcefield.

        Code inspired by:
        https://github.com/rdkit/rdkit/blob/master/Code/
        GraphMol/ForceFieldHelpers/UFF/AtomTyper.cpp

        """
        total_valence = rdkit.Atom.GetTotalValence(atom)
        atnum = int(atom.GetAtomicNum())

        # Go through element cases.
        # Mg.
        if atnum == 12:
            if total_valence == 2:
                atomkey += '+2'
            else:
                raise UFFTyperError(
                    f"UFFTYPER: Unrecognized charge state for atom: "
                    f"{atom.GetIdx}"
                )
        # Al.
        elif atnum == 13:
            if total_valence != 3:
                raise UFFTyperError(
                    f"UFFTYPER: Unrecognized charge state for atom: "
                    f"{atom.GetIdx}"
                )

        # Si.
        elif atnum == 14:
            if total_valence != 4:
                raise UFFTyperError(
                    f"UFFTYPER: Unrecognized charge state for atom: "
                    f"{atom.GetIdx}"
                )
        # P.
        elif atnum == 15:
            if total_valence == 3:
                atomkey += '+3'
            elif total_valence == 5:
                atomkey += '+5'
            else:
                raise UFFTyperError(
                    f"UFFTYPER: Unrecognized charge state for atom: "
                    f"{atom.GetIdx}"
                )

        # S.
        elif atnum == 16:
            hybrid = rdkit.Atom.GetHybridization(atom)
            if hybrid != rdkit.HybridizationType.SP2:
                if total_valence == 2:
                    atomkey += '+2'
                elif total_valence == 4:
                    atomkey += '+4'
                elif total_valence == 6:
                    atomkey += '+6'
                else:
                    raise UFFTyperError(
                        f"UFFTYPER: Unrecognized charge state for "
                        f"atom: {atom.GetIdx}"
                    )
        # Zn.
        elif atnum == 30:
            if total_valence == 2:
                atomkey += '+2'
            else:
                raise UFFTyperError(
                    f"UFFTYPER: Unrecognized charge state for atom: "
                    f"{atom.GetIdx}"
                )

        # Ga.
        elif atnum == 31:
            if total_valence == 3:
                atomkey += '+3'
            else:
                raise UFFTyperError(
                    f"UFFTYPER: Unrecognized charge state for atom: "
                    f"{atom.GetIdx}"
                )
        # As.
        elif atnum == 33:
            if total_valence == 3:
                atomkey += '+3'
            else:
                raise UFFTyperError(
                    f"UFFTYPER: Unrecognized charge state for atom: "
                    f"{atom.GetIdx}"
                )
        # Se.
        elif atnum == 34:
            if total_valence == 2:
                atomkey += '+2'
            else:
                raise UFFTyperError(
                    f"UFFTYPER: Unrecognized charge state for atom: "
                    f"{atom.GetIdx}"
                )

        # Cd.
        elif atnum == 48:
            if total_valence == 2:
                atomkey += '+2'
            else:
                raise UFFTyperError(
                    f"UFFTYPER: Unrecognized charge state for atom: "
                    f"{atom.GetIdx}"
                )
        # In.
        elif atnum == 49:
            if total_valence == 3:
                atomkey += '+3'
            else:
                raise UFFTyperError(
                    f"UFFTYPER: Unrecognized charge state for atom: "
                    f"{atom.GetIdx}"
                )

        # Sb.
        elif atnum == 51:
            if total_valence == 3:
                atomkey += '+3'
            else:
                raise UFFTyperError(
                    f"UFFTYPER: Unrecognized charge state for atom: "
                    f"{atom.GetIdx}"
                )
        # Te.
        elif atnum == 52:
            if total_valence == 2:
                atomkey += '+2'
            else:
                raise UFFTyperError(
                    f"UFFTYPER: Unrecognized charge state for atom: "
                    f"{atom.GetIdx}"
                )
        # Hg.
        elif atnum == 80:
            if total_valence == 2:
                atomkey += '+2'
            else:
                raise UFFTyperError(
                    f"UFFTYPER: Unrecognized charge state for atom: "
                    f"{atom.GetIdx}"
                )
        # Tl.
        elif atnum == 81:
            if total_valence == 3:
                atomkey += '+3'
            else:
                raise UFFTyperError(
                    f"UFFTYPER: Unrecognized charge state for atom: "
                    f"{atom.GetIdx}"
                )
        # Pb.
        elif atnum == 82:
            if total_valence == 3:
                atomkey += '+3'
            else:
                raise UFFTyperError(
                    f"UFFTYPER: Unrecognized charge state for atom: "
                    f"{atom.GetIdx}"
                )
        # Bi.
        elif atnum == 83:
            if total_valence == 3:
                atomkey += '+3'
            else:
                raise UFFTyperError(
                    f"UFFTYPER: Unrecognized charge state for atom: "
                    f"{atom.GetIdx}"
                )
        # Po.
        elif atnum == 84:
            if total_valence == 2:
                atomkey += '+2'
            else:
                raise UFFTyperError(
                    f"UFFTYPER: Unrecognized charge state for atom: "
                    f"{atom.GetIdx}"
                )
        # Lanthanides.
        elif atnum >= 57 and atnum <= 71:
            if total_valence == 6:
                atomkey += '+3'
            else:
                raise UFFTyperError(
                    f"UFFTYPER: Unrecognized charge state for atom: "
                    f"{atom.GetIdx}"
                )
        return atomkey

    def _get_atom_label(self, atom):
        """
        Get FF atom label.

        Code inspired by:
        https://github.com/rdkit/rdkit/blob/master/Code/
        GraphMol/ForceFieldHelpers/UFF/AtomTyper.cpp

        """
        atnum = int(atom.GetAtomicNum())
        atomkey = atom.GetSymbol()
        if len(atomkey) == 1:
            atomkey += '_'

        table = rdkit.GetPeriodicTable()

        chk1 = (
            rdkit.PeriodicTable.GetDefaultValence(table, atnum) == -1
        )
        chk2 = (rdkit.PeriodicTable.GetNOuterElecs(table, atnum) != 1)
        chk3 = (rdkit.PeriodicTable.GetNOuterElecs(table, atnum) != 7)
        chk4 = chk2 and chk3
        if chk1 or chk4:
            hybrid = rdkit.Atom.GetHybridization(atom)
            if atnum == 84:
                atomkey += '3'
                if hybrid != rdkit.HybridizationType.SP3:
                    raise UFFTyperError(
                        f"UFFTYPER: Unrecognized hybridization for"
                        f" atom: {atom.GetIdx}"
                    )
            elif atnum == 80:
                atomkey += '1'
                if hybrid != rdkit.HybridizationType.SP:
                    raise UFFTyperError(
                        f"UFFTYPER: Unrecognized hybridization for"
                        f" atom: {atom.GetIdx}"
                    )
            else:
                if hybrid == rdkit.HybridizationType.SP:
                    atomkey += '1'
                elif hybrid == rdkit.HybridizationType.SP2:
                    chk1a = rdkit.Atom.GetIsAromatic(atom)
                    bonds = rdkit.Atom.GetBonds(atom)
                    conjugated = False
                    for bond in bonds:
                        if rdkit.Bond.GetIsConjugated(bond):
                            conjugated = True
                            break
                    chk2a = conjugated
                    chk3a = atnum in [6, 7, 8, 16]
                    chk4a = (chk1a or chk2a)
                    if chk4a and chk3a:
                        atomkey += 'R'
                    else:
                        atomkey += '2'
                elif hybrid == rdkit.HybridizationType.SP3:
                    atomkey += '3'
                elif hybrid == rdkit.HybridizationType.SP3D:
                    atomkey += '5'
                elif hybrid == rdkit.HybridizationType.SP3D2:
                    atomkey += '6'
                else:
                    raise UFFTyperError(
                        f"UFFTYPER: Unrecognized hybridization for"
                        f" atom: {atom.GetIdx}"
                    )
        atomkey = self._add_atom_charge_flags(atom, atomkey)
        return atomkey

    def _type_translator(self):
        type_translator = {}
        types = sorted(
            set([self.atom_labels[i][0] for i in self.atom_labels])
        )
        for t in types:
            if not t[1].isalpha():
                symb = t[0]
            else:
                symb = t[0:2]
            for i in range(1, 100):
                name = f'{symb}{i}'
                if name in type_translator.values():
                    continue
                else:
                    type_translator[t] = name
                    break

        return type_translator

    def _cell_section(self, unit_cell):
        cell_section = (
            '\ncell\n'
            f"{round(unit_cell.get_a(), 6)} "
            f"{round(unit_cell.get_b(), 6)} "
            f"{round(unit_cell.get_c(), 6)} "
            f"{round(unit_cell.get_alpha(), 6)} "
            f"{round(unit_cell.get_beta(), 6)} "
            f"{round(unit_cell.get_gamma(), 6)} "
            # No fixes.
            "0 0 0 0 0 0\n"
        )

        return cell_section

    def _position_section(self, mol, type_translator):
        position_section = '\ncartesian\n'
        for atom in mol.get_atoms():
            atom_type = type_translator[self.atom_labels[
                atom.get_id()
            ][0]]
            position = mol.get_centroid(atom_ids=atom.get_id())
            posi_string = (
                f'{atom_type} core {round(position[0], 5)} '
                f'{round(position[1], 5)} {round(position[2], 5)}\n'
            )
            position_section += posi_string

        return position_section

    def _bond_section(self, mol, metal_atoms):
        bond_section = '\n'
        for bond in mol.get_bonds():
            atom_types = [
                self.atom_labels[i.get_id()][0]
                for i in [bond.get_atom1(), bond.get_atom2()]
            ]

            # Set bond orders.
            if has_h_atom(bond):
                # H has bond order of 1.
                bond_type = ''
            elif has_metal_atom(bond, metal_atoms):
                bond_type = self._metal_ligand_bond_order
            elif '_R' in atom_types[0] and '_R' in atom_types[1]:
                bond_type = 'resonant'
            elif bond.get_order() == 1:
                bond_type = ''
            elif bond.get_order() == 2:
                bond_type = 'double'
            elif bond.get_order() == 3:
                bond_type = 'triple'

            string = (
                f'connect {bond.get_atom1().get_id()+1} '
                f'{bond.get_atom2().get_id()+1} {bond_type}'
            )
            bond_section += string+'\n'

        return bond_section

    def _species_section(self, type_translator):
        species_section = '\nspecies\n'
        for spec in type_translator:
            name = type_translator[spec]
            species_section += f'{name} {spec}\n'

        return species_section

    def _write_gulp_file(
        self,
        mol,
        metal_atoms,
        in_file,
        output_xyz,
        unit_cell=None,
    ):

        type_translator = self._type_translator()

        top_line = 'opti '

        if self._conjugate_gradient:
            top_line += 'conj unit '

        if unit_cell is not None:
            # Constant pressure.
            top_line += 'conp '
            cell_section = self._cell_section(unit_cell)
            # Output CIF.
            output_cif = output_xyz.replace('xyz', 'cif')
            periodic_output = f'output cif {output_cif}\n'
        else:
            # Constant volume.
            top_line += 'conv '
            cell_section = ''
            periodic_output = ''

        top_line += 'noautobond fix molmec cartesian\n'

        position_section = self._position_section(mol, type_translator)
        bond_section = self._bond_section(mol, metal_atoms)
        species_section = self._species_section(type_translator)

        library = '\nlibrary uff4mof.lib\n'

        output_section = (
            '\n'
            f'maxcyc {self._maxcyc}\n'
            'terse inout potentials\n'
            'terse in cell\n'
            'terse in structure\n'
            'terse inout derivatives\n'
            f'output xyz {output_xyz}\n'
            f'{periodic_output}'
            # 'output movie xyz steps_.xyz\n'
        )

        with open(in_file, 'w') as f:
            f.write(top_line)
            f.write(cell_section)
            f.write(position_section)
            f.write(bond_section)
            f.write(species_section)
            f.write(library)
            f.write(output_section)

    def assign_FF(self, mol):
        """
        Assign forcefield types to molecule.

        Parameters
        ----------
        mol : :class:`stk.Molecule`
            The molecule to be optimized.

        Returns
        -------
        None : :class:`NoneType`

        """

        FutureWarning(
            'We have found some minor discrepancies in this '
            'assignment algorithm, which is based off rdkit code. '
            'Changes should come soon. This UFF optimisation should '
            ' not be your final step! Due to this, some tests in '
            'test_uff_assign_ff.py have been muted.'
        )

        metal_atoms = get_metal_atoms(mol)
        metal_ids = [i.get_id() for i in metal_atoms]

        if len(metal_ids) > 1 and self._metal_FF is None:
            raise ExpectedMetal(
                'No metal FF provivded, but metal atoms were found ('
                f'{metal_atoms})'
            )

        metal_bonds, _ = get_metal_bonds(mol, metal_atoms)
        edit_mol = to_rdkit_mol_without_metals(
            mol=mol,
            metal_atoms=metal_atoms,
            metal_bonds=metal_bonds
        )

        # Get forcefield parameters.
        rdkit.SanitizeMol(edit_mol)
        self.atom_labels = {}

        for i in range(edit_mol.GetNumAtoms()):
            if i in metal_ids:
                self.atom_labels[i] = [None, 'metal', None]
            else:
                atom = edit_mol.GetAtomWithIdx(i)
                atom_label = self._get_atom_label(atom)
                self.atom_labels[i] = [atom_label, None, None]

        # Write UFF4MOF specific forcefield parameters.
        # Metals.
        for atomid in self.atom_labels:
            if self.atom_labels[atomid][1] == 'metal':
                atom, = mol.get_atoms(atomid)
                atom_no = atom.get_atomic_number()
                self.atom_labels[atomid][0] = self._metal_FF[atom_no]

    def _run_gulp(self, in_file, out_file):
        cmd = f'{self._gulp_path} < {in_file}'
        with open(out_file, 'w') as f:
            # Note that sp.call will hold the program until completion
            # of the calculation.
            sp.call(
                cmd,
                stdin=sp.PIPE,
                stdout=f,
                stderr=sp.PIPE,
                # Shell is required to run complex arguments.
                shell=True
            )

    def extract_final_energy(self, file):
        nums = re.compile(r"[+-]?\d+(?:\.\d+)?(?:[eE][+-]?\d+)?")
        with open(file, 'r') as f:
            for line in f.readlines():
                if 'Final energy =' in line:
                    string = nums.search(line.rstrip()).group(0)
                    return float(string)

    def optimize(self, mol):
        """
        Optimize `mol`.

        Parameters
        ----------
        mol : :class:`stk.Molecule`
            The molecule to be optimized.

        Returns
        -------
        mol : :class:`.Molecule`
            The optimized molecule.

        """

        if self._output_dir is None:
            output_dir = str(uuid.uuid4().int)
        else:
            output_dir = self._output_dir
        output_dir = os.path.abspath(output_dir)

        if os.path.exists(output_dir):
            shutil.rmtree(output_dir)

        os.mkdir(output_dir)
        init_dir = os.getcwd()
        os.chdir(output_dir)

        in_file = 'gulp_opt.gin'
        out_file = 'gulp_opt.ginout'
        output_xyz = 'gulp_opt.xyz'

        metal_atoms = get_metal_atoms(mol)

        try:
            # Write GULP file.
            self._write_gulp_file(
                mol=mol,
                metal_atoms=metal_atoms,
                in_file=in_file,
                output_xyz=output_xyz,
                unit_cell=None,
            )
            # Run.
            self._run_gulp(in_file, out_file)

            # Update from output.
            mol = mol.with_structure_from_file(output_xyz)

        finally:
            os.chdir(init_dir)

        return mol

    def p_optimize(self, mol, unit_cell):
        """
        Optimize `mol` and `unit_cell`.

        Parameters
        ----------
        mol : :class:`stk.Molecule`
            The molecule to be optimized.

        unit_cell : :class:`.UnitCell`
            The unit_cell to be optimized if optimization is periodic.

        Returns
        -------
        mol : :class:`.Molecule`
            The optimized molecule.

        unit_cell : :class:`.UnitCell`
            The optimized cell.

        """

        if self._output_dir is None:
            output_dir = str(uuid.uuid4().int)
        else:
            output_dir = self._output_dir
        output_dir = os.path.abspath(output_dir)

        if os.path.exists(output_dir):
            shutil.rmtree(output_dir)

        os.mkdir(output_dir)
        init_dir = os.getcwd()
        os.chdir(output_dir)

        in_file = 'gulp_opt.gin'
        out_file = 'gulp_opt.ginout'
        output_xyz = 'gulp_opt.xyz'
        output_cif = output_xyz.replace('xyz', 'cif')

        metal_atoms = get_metal_atoms(mol)

        try:
            # Write GULP file.
            self._write_gulp_file(
                mol=mol,
                metal_atoms=metal_atoms,
                in_file=in_file,
                output_xyz=output_xyz,
                unit_cell=unit_cell,
            )

            # Run.
            self._run_gulp(in_file, out_file)

            # Update from output.
            mol = mol.with_structure_from_file(output_xyz)
            unit_cell = unit_cell.with_cell_from_cif(output_cif)

        finally:
            os.chdir(init_dir)

        return mol, unit_cell


class GulpUFFMDOptimizer(GulpUFFOptimizer):
    """
    Applies forcefield MD that can handle metal centres.

    Notes
    -----
    By default, :meth:`optimize` will run a MD run using the UFF4MOF.
    This forcefield requires some explicit metal atom definitions,
    which are determined by the user.

    """

    def __init__(
        self,
        gulp_path,
        metal_FF=None,
        metal_ligand_bond_order=None,
        output_dir=None,
        integrator='stochastic',
        ensemble='nvt',
        temperature=300,
        equilbration=1.0,
        production=10.0,
        timestep=1.0,
        N_conformers=10,
        opt_conformers=True,
        save_conformers=False,
    ):
        """
        Initialize a :class:`GulpUFFOptimizer` instance.

        Parameters
        ----------
        gulp_path : :class:`str`
            Path to GULP executable.

        metal_FF : :class:`dict`, optional
            Dictionary with metal atom forcefield assignments.
            Key: :class:`int` : atomic number.
            Value: :class:`str` : UFF4MOF forcefield type.

        metal_ligand_bond_order : :class:`str`, optional
            Bond order to use for metal-ligand bonds. Defaults to
            `half`, but using `resonant` can increase the force
            constant for stronger metal-ligand interactions.

        output_dir : :class:`str`, optional
            The name of the directory into which files generated during
            the calculation are written, if ``None`` then
            :func:`uuid.uuid4` is used.

        integrator : :class:`str`, optional
            Integrator for GULP to use.
            Defaults to 'stochastic'.

        ensemble : :class:`str`, optional
            Ensemble for GULP to use.
            Defaults to 'nvt'.

        temperature : :class:`float`, optional
            Temperature to run simulation at in Kelvin.
            Defaults to 300.

        equilbration : :class:`float`, optional
            Time spent equilibrating system in ps.
            Defaults to 1.0.

        production : :class:`float`, optional
            Time spent running production simulation of system in ps.
            Defaults to 10.0.

        timestep : :class:`float`, optional
            Timestep of simulation in fs.
            Defaults to 1.0.

        N_conformers : :class:`int`, optional
            Number of conformers to sample.
            Defaults to 10.

        opt_conformers : :class:`bool`, optional
            Whether or not to optimise each conformer using UFF4MOF.
            Defaults to ``True``.

        save_conformers : :class:`bool`, optional
            Whether or not to save to file each conformer.
            Defaults to ``False``.

        """
        self._gulp_path = gulp_path
        self._metal_FF = metal_FF
        self._metal_ligand_bond_order = (
            'half' if metal_ligand_bond_order is None
            else metal_ligand_bond_order
        )
        self._output_dir = output_dir
        self._integrator = integrator
        self._ensemble = ensemble
        self._temperature = temperature
        self._equilbration = float(equilbration)
        self._production = float(production)
        self._timestep = timestep
        self._N_conformers = N_conformers
        samples = float(self._production) / float(self._N_conformers)
        self._sample = samples
        self._write = samples
        self._opt_conformers = opt_conformers
        self._save_conformers = save_conformers

    def _write_gulp_file(
        self,
        mol,
        metal_atoms,
        in_file,
        output_traj
    ):

        type_translator = self._type_translator()

        top_line = 'md conv noautobond fix molmec cartesian\n'

        position_section = self._position_section(mol, type_translator)
        bond_section = self._bond_section(mol, metal_atoms)
        species_section = self._species_section(type_translator)

        library = '\nlibrary uff4mof.lib\n'

        md_section = (
            '\n'
            "mdmaxtemp 100000000\n"
            f"integrator {self._integrator}\n"
            f"ensemble {self._ensemble}\n"
            f"temperature {self._temperature}\n"
            f"equilbration {self._equilbration} ps\n"
            f"production {self._production} ps\n"
            f"timestep {self._timestep} fs\n"
            f"sample {self._sample} ps\n"
            f"write {self._write} ps\n"
            '\n'
        )

        output_section = (
            '\n'
            'terse inout potentials\n'
            'terse inout cell\n'
            'terse in structure\n'
            'terse inout derivatives\n'
            f'output trajectory ascii {output_traj}\n'
        )

        with open(in_file, 'w') as f:
            f.write(top_line)
            f.write(position_section)
            f.write(bond_section)
            f.write(species_section)
            f.write(library)
            f.write(md_section)
            f.write(output_section)

    def _convert_traj_to_xyz(self, output_xyz, output_traj):

        # Get atom types from an existing xyz file.
        atom_types = []
        with open(output_xyz, 'r') as f:
            for line in f.readlines()[2:]:
                atom_types.append(line.rstrip().split(' ')[0])

        # Read in lines from trajectory file.
        with open(output_traj, 'r') as f:
            lines = f.readlines()

        # Split file using strings.
        timesteps = ''.join(lines).split('#  Time/KE/E/T')[1:]
        trajectory_data = {}
        xyz_traj_lines = []
        for ts, cont in enumerate(timesteps):
            ts_data = {}
            time_section = (
                cont.split('#  Coordinates\n')[0]
            )
            coords_section = (
                cont.split('#  Coordinates\n')[1].split(
                    '#  Velocities\n'
                )[0]
            )
            vels_section = (
                cont.split('#  Velocities\n')[1].split(
                    '#  Derivatives \n'
                )[0]
            )
            derivs_section = (
                cont.split('#  Derivatives \n')[1].split(
                    '#  Site energies \n'
                )[0]
            )
            sites_section = (
                cont.split('#  Site energies \n')[1]
            )

            ts_data['time'] = float(
                [i for i in time_section.strip().split(' ') if i][0]
            )
            ts_data['KE'] = float(
                [i for i in time_section.strip().split(' ') if i][1]
            )
            ts_data['E'] = float(
                [i for i in time_section.strip().split(' ') if i][2]
            )
            ts_data['T'] = float(
                [i for i in time_section.strip().split(' ') if i][3]
            )
            ts_data['coords'] = [
                [i for i in li.split(' ') if i]
                for li in coords_section.split('\n')[:-1]
            ]
            ts_data['vels'] = [
                [i for i in li.split(' ') if i]
                for li in vels_section.split('\n')[:-1]
            ]
            ts_data['derivs'] = [
                [i for i in li.split(' ') if i]
                for li in derivs_section.split('\n')[:-1]
            ]
            ts_data['sites'] = [
                [i for i in li.split(' ') if i]
                for li in sites_section.split('\n')[:-1]
            ]

            trajectory_data[ts] = ts_data

            # Write XYZ string for XYZ traj file.
            xyz_string = (
                f"{len(ts_data['coords'])}\n"
                f"{ts_data['time']},{ts_data['KE']},"
                f"{ts_data['E']},{ts_data['T']}\n"
            )

            for i, coord in enumerate(ts_data['coords']):
                site_E = ts_data['sites'][i][0]
                xyz_string += (
                    f'{atom_types[i]} {round(float(coord[0]), 5)} '
                    f'{round(float(coord[1]), 5)} '
                    f'{round(float(coord[2]), 5)} {site_E}\n'
                )

            xyz_traj_lines.append(xyz_string)

        return atom_types, trajectory_data, xyz_traj_lines

    def _write_conformer_xyz_file(
        self,
        ts,
        ts_data,
        filename,
        atom_types,
    ):

        coords = ts_data['coords']
        xyz_string = (
            f"{len(ts_data['coords'])}\n"
            f"{ts},{ts_data['time']},{ts_data['KE']},"
            f"{ts_data['E']},{ts_data['T']}\n"
        )

        for i, coord in enumerate(coords):
            xyz_string += (
                f'{atom_types[i]} {round(float(coord[0]), 5)} '
                f'{round(float(coord[1]), 5)} '
                f'{round(float(coord[2]), 5)}\n'
            )
        with open(filename, 'w') as f:
            f.write(xyz_string)

    def _optimise_all_conformers(
        self,
        mol,
        trajectory_data,
        atom_types,
        low_conf_xyz,
    ):
        min_energy = 1E10
        for ts in trajectory_data:
            if self._save_conformers:
                conformer_file_name = f'conf_{ts}.xyz'
            else:
                # This will get overwrriten each time and deleted at
                # the end.
                conformer_file_name = 'temp_conf.xyz'

            self._write_conformer_xyz_file(
                ts=ts,
                ts_data=trajectory_data[ts],
                filename=conformer_file_name,
                atom_types=atom_types
            )
            mol = mol.with_structure_from_file(conformer_file_name)
            conformer_opt_dir = os.path.join(
                os.getcwd(),
                f'conf_{ts}_opt'
            )
            gulp_opt = GulpUFFOptimizer(
                gulp_path=self._gulp_path,
                metal_FF=self._metal_FF,
                output_dir=conformer_opt_dir,
            )
            gulp_opt.assign_FF(mol)
            mol = gulp_opt.optimize(mol=mol)
            energy = gulp_opt.extract_final_energy(
                file=os.path.join(
                    conformer_opt_dir,
                    'gulp_opt.ginout'
                )
            )
            if energy < min_energy:
                min_energy = energy
                # Write out optimised conformer.
                mol.write(low_conf_xyz)

        if not self._save_conformers:
            os.remove('temp_conf.xyz')

    def _save_all_conformers(self, trajectory_data, atom_types):
        for ts in trajectory_data:
            conformer_file_name = f'conf_{ts}.xyz'
            self._write_conformer_xyz_file(
                ts=ts,
                ts_data=trajectory_data[ts],
                filename=conformer_file_name,
                atom_types=atom_types
            )

    def _calculate_lowest_energy_conformer(self, trajectory_data):
        energies = [
            trajectory_data[ts]['E']
            for ts in trajectory_data
        ]
        min_energy = min(energies)
        min_ts = list(trajectory_data.keys())[
            energies.index(min_energy)
        ]
        return min_ts

    def _save_lowest_energy_conf(
        self,
        mol,
        output_xyz,
        output_traj,
        xyz_traj,
        low_conf_xyz
    ):

        # Convert GULP trajectory file to xyz trajectory.
        atom_types, trajectory_data, xyz_traj_lines = (
            self._convert_traj_to_xyz(
                output_xyz=output_xyz,
                output_traj=output_traj,
            )
        )
        # Write XYZ trajectory file.
        with open(xyz_traj, 'w') as f:
            for line in xyz_traj_lines:
                f.write(line)

        # Find lowest energy conformation and output to XYZ.
        if self._opt_conformers:
            self._optimise_all_conformers(
                mol=mol,
                trajectory_data=trajectory_data,
                atom_types=atom_types,
                low_conf_xyz=low_conf_xyz,
            )
        else:
            if self._save_conformers:
                self._save_all_conformers(trajectory_data, atom_types)

            min_ts = self._calculate_lowest_energy_conformer(
                trajectory_data=trajectory_data,
            )

            self._write_conformer_xyz_file(
                ts=min_ts,
                ts_data=trajectory_data[min_ts],
                filename=low_conf_xyz,
                atom_types=atom_types
            )

    def optimize(self, mol):
        """
        Optimize `mol`.

        Parameters
        ----------
        mol : :class:`stk.Molecule`
            The molecule to be optimized.

        Returns
        -------
        mol : :class:`stk.Molecule`
            The optimized molecule.

        """

        if self._output_dir is None:
            output_dir = str(uuid.uuid4().int)
        else:
            output_dir = self._output_dir
        output_dir = os.path.abspath(output_dir)

        if os.path.exists(output_dir):
            shutil.rmtree(output_dir)

        os.mkdir(output_dir)
        init_dir = os.getcwd()
        os.chdir(output_dir)

        in_file = 'gulp_MD.gin'
        out_file = 'gulp_MD.ginout'
        output_xyz = 'gulp_MD_template.xyz'
        output_traj = 'gulp_MD.trg'
        xyz_traj = 'gulp_MD_traj.xyz'
        low_conf_xyz = 'low_energy_conf.xyz'

        metal_atoms = get_metal_atoms(mol)

        try:
            mol.write(output_xyz)
            # Write GULP file.
            self._write_gulp_file(
                mol=mol,
                metal_atoms=metal_atoms,
                in_file=in_file,
                output_traj=output_traj
            )

            # Run.
            self._run_gulp(in_file, out_file)

            # Get lowest energy conformer from trajectory.
            self._save_lowest_energy_conf(
                mol=mol,
                output_xyz=output_xyz,
                output_traj=output_traj,
                xyz_traj=xyz_traj,
                low_conf_xyz=low_conf_xyz
            )

            # Update from output.
            mol = mol.with_structure_from_file(low_conf_xyz)

        finally:
            os.chdir(init_dir)

        return mol



"""
XTB Extractor
=============

#. :class:`.XTBExtractor`

Class to extract properties from xTB output.

"""

import re

from .extractor import Extractor


class XTBExtractor(Extractor):
    """
    Extracts properties from xTB output files.

    All formatting based on the 190418 version of xTB.

    Attributes
    ----------
    output_file : :class:`str`
        Output file to extract properties from.

    output_lines : :class:`list` : :class:`str`
        :class:`list` of all lines in as :class:`str` in the output
        file.

    total_energy : :class:`float`
        The total energy in the :attr:`output_file` as
        :class:`float`. The energy is in units of a.u..

    homo_lumo_gap : :class:`float`
        The HOMO-LUMO gap in the :attr:`output_file` as
        :class:`float`. The gap is in units of eV.

    fermi_level : :class:`float`
        The Fermi level in the :attr:`output_file` as
        :class:`float` in units of eV.

    qonly_dipole_moment : :class:`list`
        Components of the Q only dipole moment in units
        of Debye in :class:`list` of the form
        ``[x, y, z]``.

    full_dipole_moment : :class:`list`
        Components of the full dipole moment in units
        of Debye in :class:`list` of the form
        ``[x, y, z, total]``.

    qonly_quadrupole_moment : :class:`list`
        Components of the Q only traceless quadrupole moment in units
        of Debye in :class:`list` of the form
        ``[xx, xy, xy, xz, yz, zz]``.

    qdip_quadrupole_moment : :class:`list`
        Components of the Q+Dip traceless quadrupole moment in units of
        Debye in :class:`list` of the form
        ``[xx, xy, xy, xz, yz, zz]``.

    full_quadrupole_moment : :class:`list`
        Components of the full traceless quadrupole moment in units of
        Debye in :class:`list` of the form
        ``[xx, xy, xy, xz, yz, zz]``.

    homo_lumo_occ : :class:`dict`
        :class:`dict` of :class:`list` containing the orbital number,
        energy in eV and occupation of the HOMO and LUMO orbitals in
        the :attr:`output_file`.

    total_free_energy : :class:`float`
        The total free energy in the :attr:`output_file` as
        :class:`float`. The free energy is in units of a.u. and
        calculated at 298.15K.

    frequencies : :class:`list`
        :class:`list` of the vibrational frequencies as :class:`float`
        in the :attr:`output_file`. Vibrational frequencies are in
        units of wavenumber and calculated at 298.15K.

    ionisation_potential : :class:`float`
        The vertical ionisation potential in the :attr:`output_file` as
        :class:`float`. Corresponds to the delta SCC IP.

    electron_affinity : :class:`float`
        The vertical electron affinity in the :attr:`output_file` as
        :class:`float`. Corresponds to the delta SCC EA.

    """

    def _extract_values(self):

        for i, line in enumerate(self.output_lines):
            if self._check_line(line, 'total_energy'):
                self._extract_total_energy(line)
            elif self._check_line(line, 'homo_lumo_gap'):
                self._extract_homo_lumo_gap(line)
            elif self._check_line(line, 'fermi_level'):
                self._extract_fermi_level(line)
            elif self._check_line(line, 'dipole_moment'):
                self._extract_qonly_dipole_moment(i)
                self._extract_full_dipole_moment(i)
            elif self._check_line(line, 'quadrupole_moment'):
                self._extract_qonly_quadrupole_moment(i)
                self._extract_qdip_quadrupole_moment(i)
                self._extract_full_quadrupole_moment(i)
            elif self._check_line(line, 'homo_lumo_occ_HOMO'):
                self.homo_lumo_occ = {}
                self._extract_homo_lumo_occ(line, 'HOMO')
            elif self._check_line(line, 'homo_lumo_occ_LUMO'):
                self._extract_homo_lumo_occ(line, 'LUMO')
            elif self._check_line(line, 'total_free_energy'):
                self._extract_total_free_energy(line)
            elif self._check_line(line, 'ionisation_potential'):
                self._extract_ionisation_potential(line)
            elif self._check_line(line, 'electron_affinity'):
                self._extract_electron_affinity(line)

        # Frequency formatting requires loop through full file.
        self._extract_frequencies()

    def _properties_dict(self):

        return {
            'total_energy': '          | TOTAL ENERGY  ',
            'homo_lumo_gap': '          | HOMO-LUMO GAP   ',
            'fermi_level': '             Fermi-level        ',
            'dipole_moment': 'molecular dipole:',
            'quadrupole_moment': 'molecular quadrupole (traceless):',
            'homo_lumo_occ_HOMO': '(HOMO)',
            'homo_lumo_occ_LUMO': '(LUMO)',
            'total_free_energy': '          | TOTAL FREE ENERGY  ',
            'ionisation_potential': 'delta SCC IP (eV)',
            'electron_affinity': 'delta SCC EA (eV)',
        }

    def _extract_total_energy(self, line):
        """
        Updates :attr:`total_energy`.

        Parameters
        ----------
        line : :class:`str`
            Line of output file to extract property from.

        Returns
        -------
        None : :class:`NoneType`

        """

        nums = re.compile(r"[+-]?\d+(?:\.\d+)?(?:[eE][+-]?\d+)?")
        string = nums.search(line.rstrip()).group(0)
        self.total_energy = float(string)

    def _extract_homo_lumo_gap(self, line):
        """
        Updates :attr:`homo_lumo_gap`.

        Parameters
        ----------
        line : :class:`str`
            Line of output file to extract property from.

        Returns
        -------
        None : :class:`NoneType`

        """

        nums = re.compile(r"[+-]?\d+(?:\.\d+)?(?:[eE][+-]?\d+)?")
        string = nums.search(line.rstrip()).group(0)
        self.homo_lumo_gap = float(string)

    def _extract_fermi_level(self, line):
        """
        Updates :attr:`fermi_level`.

        Parameters
        ----------
        line : :class:`str`
            Line of output file to extract property from.

        Returns
        -------
        None : :class:`NoneType`

        """

        nums = re.compile(r"[+-]?\d+(?:\.\d+)?(?:[eE][+-]?\d+)?")
        part2 = line.split('Eh')
        string = nums.search(part2[1].rstrip()).group(0)
        self.fermi_level = float(string)

    def _extract_qonly_dipole_moment(self, index):
        """
        Updates :attr:`qonly_dipole_moment`.

        Parameters
        ----------
        line : :class:`str`
            Line of output file to extract property from.

        index : :class:`int`
            Index of line in :attr:`output_lines`.

        Returns
        -------
        None : :class:`NoneType`

        """

        sample_set = self.output_lines[index+2].rstrip()

        if 'q only:' in sample_set:
            self.qonly_dipole_moment = [
                float(i)
                for i in sample_set.split(':')[1].split(' ') if i
            ]

    def _extract_full_dipole_moment(self, index):
        """
        Updates :attr:`full_dipole_moment`.

        Parameters
        ----------
        line : :class:`str`
            Line of output file to extract property from.

        index : :class:`int`
            Index of line in :attr:`output_lines`.

        Returns
        -------
        None : :class:`NoneType`

        """

        sample_set = self.output_lines[index+3].rstrip()

        if 'full:' in sample_set:
            self.full_dipole_moment = [
                float(i)
                for i in sample_set.split(':')[1].split(' ') if i
            ]

    def _extract_qonly_quadrupole_moment(self, index):
        """
        Updates :attr:`qonly_quadrupole_moment`.

        Parameters
        ----------
        line : :class:`str`
            Line of output file to extract property from.

        index : :class:`int`
            Index of line in :attr:`output_lines`.

        Returns
        -------
        None : :class:`NoneType`

        """

        sample_set = self.output_lines[index+2].rstrip()

        if 'q only:' in sample_set:
            self.qonly_quadrupole_moment = [
                float(i)
                for i in sample_set.split(':')[1].split(' ') if i
            ]

    def _extract_qdip_quadrupole_moment(self, index):
        """
        Updates :attr:`qdip_quadrupole_moment`.

        Parameters
        ----------
        line : :class:`str`
            Line of output file to extract property from.

        index : :class:`int`
            Index of line in :attr:`output_lines`.

        Returns
        -------
        None : :class:`NoneType`

        """

        sample_set = self.output_lines[index+3].rstrip()

        if 'q+dip:' in sample_set:
            self.qdip_quadrupole_moment = [
                float(i)
                for i in sample_set.split(':')[1].split(' ') if i
            ]

    def _extract_full_quadrupole_moment(self, index):
        """
        Updates :attr:`full_quadrupole_moment`.

        Parameters
        ----------
        line : :class:`str`
            Line of output file to extract property from.

        index : :class:`int`
            Index of line in :attr:`output_lines`.

        Returns
        -------
        None : :class:`NoneType`

        """

        sample_set = self.output_lines[index+4].rstrip()

        if 'full:' in sample_set:
            self.full_quadrupole_moment = [
                float(i)
                for i in sample_set.split(':')[1].split(' ') if i
            ]

    def _extract_homo_lumo_occ(self, line, orbital):
        """
        Updates :attr:`homo_lumo_occ`.

        Parameters
        ----------
        line : :class:`str`
            Line of output file to extract property from.

        orbital : :class:`str`
            Can be 'HOMO' or 'LUMO'.

        Returns
        -------
        None : :class:`NoneType`

        """

        if orbital == 'HOMO':
            split_line = [i for i in line.rstrip().split(' ') if i]
            # The line is:
            #   Number, occupation, energy (Ha), energy (ev), label
            # Extract:
            #   Number, occupation, energy (eV)
            orbital_val = [
                int(split_line[0]),
                float(split_line[1]),
                float(split_line[3])
            ]
        elif orbital == 'LUMO':
            split_line = [i for i in line.rstrip().split(' ') if i]
            # The line is:
            #   Number, energy (Ha), energy (ev), label
            # Extract:
            #   Number, occupation (zero), energy (eV)
            orbital_val = [
                int(split_line[0]),
                0,
                float(split_line[2])
            ]

        self.homo_lumo_occ[orbital] = orbital_val

    def _extract_total_free_energy(self, line):
        """
        Updates :attr:`total_free_energy`.

        Returns
        -------
        None : :class:`NoneType`

        """

        nums = re.compile(r"[+-]?\d+(?:\.\d+)?(?:[eE][+-]?\d+)?")
        string = nums.search(line.rstrip()).group(0)
        self.total_free_energy = float(string)

    def _extract_frequencies(self):
        """
        Updates :attr:`frequencies`.

        Returns
        -------
        None : :class:`NoneType`

        """

        test = '|               Frequency Printout                |'

        # Use a switch to make sure we are extracting values after the
        # final property readout.
        switch = False

        frequencies = []
        for i, line in enumerate(self.output_lines):
            if test in line:
                # Turn on reading as final frequency printout has
                # begun.
                switch = True
            if ' reduced masses (amu)' in line:
                # Turn off reading as frequency section is done.
                switch = False
            if 'eigval :' in line and switch is True:
                samp = line.rstrip().split(':')[1].split(' ')
                split_line = [i for i in samp if i]
                for freq in split_line:
                    frequencies.append(freq)

        self.frequencies = [float(i) for i in frequencies]

    def _extract_ionisation_potential(self, line):
        """
        Updates :attr:`ionisation_potential`.

        Returns
        -------
        None : :class:`NoneType`

        """

        nums = re.compile(r"[+-]?\d+(?:\.\d+)?(?:[eE][+-]?\d+)?")
        string = nums.search(line.rstrip()).group(0)
        self.ionisation_potential = float(string)

    def _extract_electron_affinity(self, line):
        """
        Updates :attr:`electron_affinity`.

        Returns
        -------
        None : :class:`NoneType`

        """

        nums = re.compile(r"[+-]?\d+(?:\.\d+)?(?:[eE][+-]?\d+)?")
        string = nums.search(line.rstrip()).group(0)
        self.electron_affinity = float(string)
